from __future__ import unicode_literals, print_function

import pytest
import string
import time
from itertools import izip_longest
from random import Random

from treeoflife import searching
from treeoflife.tracker import Tracker
from treeoflife.parseutil import ParseError
from treeoflife.test.util import FakeNodeCreator, match
from treeoflife.nodes.misc import GenericActivate
from treeoflife.file_storage import serialize
from treeoflife.exceptions import LoadError


@pytest.fixture(params=[searching.parse_single, searching.parse])
def makequery(request):
    return request.param


def test_tags_parsing():
    assert searching.SearchGrammar("around, the, world").tag_texts() == (
            "around", "the", "world")
    assert searching.SearchGrammar("around the world").tag_texts() == (
            "around the world",)


def test_derp():
    assert searching.SearchGrammar(" :{test}").tags() == ("test",)
    assert searching.SearchGrammar(" :{test}").tagged_node()
    assert searching.SearchGrammar("test :{test, test}").tagged_node()
    assert repr(searching.SearchGrammar("test > test :{test}").query())
    assert repr(searching.SearchGrammar("test :{test}").query())


@pytest.mark.parametrize(("querytext", "node_type", "text"), [
    ("reference: '<< > target'", "reference", "<< > target"),

    (r"reference: '<< > \'target'", "reference", r"""<< > 'target"""),

    (r"""reference: '<< > \\"target'""", "reference", r"""<< > \"target"""),

    (r'reference: "<< > \"target\""', "reference", '<< > "target"'),

    ("""reference: '<< > target > reference: "<< > *"' """,
        "reference", '<< > target > reference: "<< > *"'),

    ('''reference: "<< > target > reference: '<< > *'" ''',
        "reference", "<< > target > reference: '<< > *'"),

    ("task: do something with 'something'",
        "task", "do something with 'something'"),

    ('task: do something with "something"',
        "task", 'do something with "something"'),

    ('''task: do 'something' with "something"''',
        "task", 'do \'something\' with "something"'),
])
def test_quoting(querytext, node_type, text):
    q = searching.parse_single(querytext)
    assert q.segments[0].pattern.type == node_type
    assert q.segments[0].pattern.text == text
    assert len(q.segments) == 1


def test_search(makequery):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    origin = tracker.root.createchild("task", "origin")
    target1 = origin.createchild("task", "target1")
    ignoreme = origin.createchild("task", "ignoreme")
    ignoreme.createchild("task", "ignore me too")
    peer = tracker.root.createchild("task", "peer")
    target2 = peer.createchild("task", "Target2")

    query1 = makequery("task: target1")
    assert query1(origin).list() == [target1]
    assert repr(query1)

    query2 = makequery("-> task: PEER > task: tarGet2")
    assert query2(origin).list() == [target2]
    assert repr(query2)


def test_id_lookup(makequery):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    origin = tracker.root.createchild("task", "origin", nodeid="orign")
    target1 = origin.createchild("task", "target1", nodeid="abcde")
    target3 = origin.createchild("task", "#abcde")
    ignoreme = origin.createchild("task", "ignoreme")
    ignoreme.createchild("task", "ignore me too")
    peer = tracker.root.createchild("task", "peer")
    target2 = peer.createchild("task", "target2", nodeid="trgt2")
    targetlist = [
        target2.createchild("task") for x in range(10)]

    query1 = makequery("#abcde")
    assert query1(origin).list() == [target1]
    assert query1(target2).list() == [target1]
    assert repr(query1)

    query2 = searching.parse_single("#orign -> task: peer > task: target2")
    assert query2(ignoreme).list() == [target2]
    assert query2(tracker.root).list() == [target2]
    assert repr(query2)

    query3 = makequery("#trgt2>")
    assert query3(ignoreme).list() == targetlist
    assert query3(tracker.root).list() == targetlist
    assert repr(query3)

    query4 = makequery("*: #abcde")
    assert query4(origin).list() == [target3]
    assert repr(query4)

    query5 = makequery("#nonex")
    assert query5(origin).first() is None
    assert repr(query4)


def test_id_lookup_invalid():
    with pytest.raises(Exception):
        print(makequery("-> #orign"))
    with pytest.raises(Exception):
        print(makequery("-> task -> #orign"))
    with pytest.raises(Exception):
        print(makequery("#orign test"))


def test_pluralities():
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    origin = tracker.root.createchild("task", "origin")

    targets = []

    for i in range(10):
        targets.append(tracker.root.createchild("task", "target %d" % i))

    query1 = searching.parse_single("-> * :{first}")
    assert query1(origin).list() == [targets[0]]
    assert repr(query1)

    query1 = searching.parse_single("-> * :{last}")
    assert query1(origin).list() == [targets[-1]]
    assert repr(query1)

    query1 = searching.parse_single("-> * :{many}")
    assert query1(origin).list() == targets
    assert repr(query1)


def test_flatten(makequery):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    origin = tracker.root.createchild("task", "origin")

    nodes = []
    targets = []

    nodes.append(origin.createchild("task", "a"))
    nodes.append(nodes[-1].createchild("task", "target"))
    targets.append(nodes[-1])
    nodes.append(origin.createchild("task", "c"))
    nodes.append(nodes[-1].createchild("task", "target"))
    targets.append(nodes[-1])
    nodes.append(nodes[-1].createchild("task", "e"))
    nodes.append(nodes[-2].createchild("task", "target"))
    targets.append(nodes[-1])
    nodes.append(nodes[-1].createchild("task", "g"))
    nodes.append(origin.createchild("task", "target"))
    targets.append(nodes[-1])
    nodes.append(nodes[-1].createchild("task", "target"))
    targets.append(nodes[-1])
    nodes.append(nodes[-1].createchild("task", "target"))
    targets.append(nodes[-1])
    nodes.append(nodes[-2].createchild("task", "k"))

    query1 = makequery("**")

    assert query1(origin).list() == nodes
    assert repr(query1)

    query2 = makequery("** > task: target")
    a = query2(origin).list()
    assert a == targets
    assert repr(query2)


def test_flat_text():
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    origin = tracker.root.createchild("task", "origin")

    targets = []
    target1 = origin.createchild("task", "target1")
    targets.append(target1)
    origin.createchild("comment", "ignoreme")
    targets.append(origin.createchild("task", "don't ignore me"))
    origin.createchild("comment", "legitimately ignore me")
    target2 = origin.createchild("task", "target2")
    targets.append(target2)

    query1 = searching.parse_single("task")
    assert query1(origin).list() == targets
    assert repr(query1)

    query2 = searching.parse_single("target1")
    assert query2(origin).list() == [target1]
    assert repr(query2)

    query3 = searching.parse_single("target2")
    assert query3(origin).list() == [target2]
    assert repr(query3)


def test_prev_peer():
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    target1 = tracker.root.createchild("task", "target1")
    tracker.root.createchild("comment", "ignoreme")
    tracker.root.createchild("task", "don't ignore me")
    tracker.root.createchild("comment", "legitimately ignore me")
    target2 = tracker.root.createchild("task", "target2")
    origin = tracker.root.createchild("task", "origin")

    query1 = searching.parse_single("<- target1")
    assert query1(origin).list() == [target1]
    assert repr(query1)

    query2 = searching.parse_single("<- target2")
    assert query2(origin).list() == [target2]
    assert repr(query2)


def test_parents():
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    parent3 = tracker.root.createchild("task", "parent3")
    parent2 = parent3.createchild("task", "parent2")
    parent1 = parent2.createchild("task", "parent1")
    origin1 = parent1.createchild("task", "origin1")

    expected = [
        parent1, parent2, parent3, tracker.root
    ]

    query = searching.parse_single("<")
    assert query(origin1).list() == expected
    assert repr(query)


def test_tags_filtering():
    class CustomGeneric(GenericActivate):
        def search_tags(self):
            return self.text.split()

    tracker = Tracker(False, FakeNodeCreator(CustomGeneric))

    targets = []
    nontargets = []

    origin1 = tracker.root.createchild("task", "origin")
    targets.append(origin1.createchild("task", "target node"))
    nontargets.append(origin1.createchild("task", "nontarget node"))
    targets.append(origin1.createchild("task", "target node"))
    nontargets.append(origin1.createchild("task", "nontarget node"))

    query1 = searching.parse_single("task :{target}")
    assert query1(origin1).list() == targets
    assert repr(query1)

    query1 = searching.parse_single("task :{nontarget}")
    assert query1(origin1).list() == nontargets
    assert repr(query1)


def test_node_with_colon():
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))
    origin = tracker.root.createchild("task", 'origin')
    target = tracker.root.createchild("task", 'target: with colon')

    query = searching.parse_single("-> task: target: with colon")
    assert query(origin).list() == [target]
    assert repr(query)


class TestCreate(object):
    def test_create_node(self):
        tracker = Tracker(False, FakeNodeCreator(GenericActivate))

        origin = tracker.root.createchild("task", "origin")

        creator = searching.parse_create("task: Target")
        creator(origin)

        assert match("\n".join(serialize(origin)), (
            "task#?????: origin\n"
            "    task#?????: Target"
        ))
        assert repr(creator)

    def test_create_node_existing(self):
        tracker = Tracker(False)

        nodes = []

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(origin.createchild("task", "untouched"))
        nodes.append(origin.createchild("task", "untouched 2"))

        creator = searching.parse_create("task: target")
        created = creator(origin)

        results = list(origin.children)
        assert (results[:2] + results[3:]) == nodes
        assert results[2].node_type == "task"
        assert results[2].text == "target"
        assert created == results[2]
        assert repr(creator)

    def test_existing_query(self):
        tracker = Tracker(False, FakeNodeCreator(GenericActivate))

        origin = tracker.root.createchild("task", "origin")

        selector = searching.parse_single("task: target")

        creator = searching.parse_create(query=selector)
        creator(origin)

        assert match("\n".join(serialize(origin)), (
            "task#?????: origin\n"
            "    task#?????: target"
        ))
        assert repr(creator)

    # removed this functionality

    #def test_multi_create(self):
    #    tracker = Tracker(False)

    #    nodes = []

    #    origin = tracker.root.createchild("task", "origin")
    #    nodes.append(origin.createchild("task", "finished"))
    #    nodes[-1].start()
    #    nodes[-1].finish()
    #    nodes.append(origin.createchild("task", "finished 2"))
    #    nodes[-1].start()
    #    nodes[-1].finish()
    #    nodes.append(origin.createchild("task", "started"))
    #    nodes[-1].start()
    #    nodes.append(origin.createchild("task", "untouched"))
    #    nodes.append(origin.createchild("task", "untouched 2"))

    #    selector = searching.parse_single("task: target :{many}")
    #    #import pudb; pudb.set_trace()
    #    creator = searching.parse_create(query=selector)
    #    #import pytest; pytest.set_trace()
    #    created = creator(origin)

    #    results = list(origin.children)
    #    assert [
    #        results[0],
    #        results[1],
    #        results[3],
    #        results[5],
    #        results[7],
    #    ] == nodes
    #    assert results[2].node_type == "task"
    #    assert results[2].text == "target"
    #    assert results[4].node_type == "task"
    #    assert results[4].text == "target"
    #    assert results[6].node_type == "task"
    #    assert results[6].text == "target"
    #    assert created == [results[2], results[4], results[6]]
    #    assert repr(creator)

    def test_after_create(self):
        tracker = Tracker(False, FakeNodeCreator(GenericActivate))

        origin = tracker.root.createchild("task", "origin")
        origin.createchild("task", "1")
        origin.createchild("task", "2")
        origin.createchild("task", "3")
        origin.createchild("task", "4")
        origin.createchild("task", "5")

        creator = searching.parse_create_single("task: target :{first, after}")
        creator(origin)

        assert match("\n".join(serialize(origin)), (
            "task#?????: origin\n"
            "    task#?????: 1\n"
            "    task#?????: target\n"
            "    task#?????: 2\n"
            "    task#?????: 3\n"
            "    task#?????: 4\n"
            "    task#?????: 5"
        ))
        assert repr(creator)

    def test_after_started(self):
        tracker = Tracker(False)

        nodes = []

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(origin.createchild("task", "untouched"))
        nodes.append(origin.createchild("task", "untouched 2"))

        creator = searching.parse_create_single(
                "task: target :{after, last, finished}")
        created = creator(origin)

        results = list(origin.children)
        assert (results[:2] + results[3:]) == nodes
        assert results[2].node_type == "task"
        assert results[2].text == "target"
        assert created == results[2]
        assert repr(creator)

    def test_last(self):
        creator = searching.parse_create_single("+task: last")
        assert not creator.is_before
        assert creator.last_segment.tags == set()
        assert creator.last_segment.plurality == "last"

        tracker = Tracker(False)

        nodes = []
        origin = tracker.root.createchild("task", "origin")

        nodes.append(origin.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(origin.createchild("task", "untouched"))
        nodes.append(origin.createchild("task", "untouched 2"))

        creator(origin)

        results = list(origin.children)

        assert results[-1].node_type == "task"
        assert results[-1].text == "last"
        assert results[:-1] == nodes
        assert repr(creator)

    def test_early(self):
        creator = searching.parse_create_single("-task: first")
        assert creator.is_before
        assert creator.last_segment.tags == set()
        assert creator.last_segment.plurality == "first"

        tracker = Tracker(False)

        nodes = []
        origin = tracker.root.createchild("task", "origin")

        nodes.append(origin.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(origin.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(origin.createchild("task", "untouched"))
        nodes.append(origin.createchild("task", "untouched 2"))

        creator(origin)

        results = list(origin.children)

        assert results[0].node_type == "task"
        assert results[0].text == "first"
        assert results[1:] == nodes
        assert repr(creator)

    def test_early_reversed(self):
        creator = searching.parse_create_single("<- +task: first")
        assert creator.is_before
        assert creator.last_segment.tags == set()
        assert creator.last_segment.plurality == "last"

        tracker = Tracker(False)

        nodes = []

        nodes.append(tracker.root.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(tracker.root.createchild("task", "untouched"))
        nodes.append(tracker.root.createchild("task", "untouched 2"))

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin)

        creator(origin)

        results = list(tracker.root.children)

        assert results[0].node_type == "task"
        assert results[0].text == "first"
        assert results[1:] == nodes
        assert repr(creator)

    def test_last_reversed(self):
        creator = searching.parse_create_single("<- -task: first")
        assert not creator.is_before
        assert creator.last_segment.tags == set()
        assert creator.last_segment.plurality == "first"

        tracker = Tracker(False)

        nodes = []

        nodes.append(tracker.root.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(tracker.root.createchild("task", "untouched"))
        nodes.append(tracker.root.createchild("task", "untouched 2"))

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin)

        creator(origin)

        results = list(tracker.root.children)

        assert results[-2].node_type == "task"
        assert results[-2].text == "first"
        assert results[:-2] + [origin] == nodes
        assert repr(creator)

    def test_default_reversed(self):
        creator = searching.parse_create_single("<- task: mid")
        assert not creator.is_before
        assert creator.last_segment.tags == set(["can_activate"])
        assert creator.last_segment.plurality == "first"

        tracker = Tracker(False)

        nodes = []

        nodes.append(tracker.root.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(tracker.root.createchild("task", "untouched"))
        nodes.append(tracker.root.createchild("task", "untouched 2"))
        nodes.append(tracker.root.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin)

        creator(origin)

        results = list(tracker.root.children)

        assert results[-3].node_type == "task"
        assert results[-3].text == "mid"
        assert results[:-3] + results[-2:] == nodes
        assert repr(creator)

    def test_reversed_no_match(self):
        creator = searching.parse_create_single("<- task: mid")
        assert not creator.is_before
        assert creator.last_segment.tags == set(["can_activate"])
        assert creator.last_segment.plurality == "first"

        tracker = Tracker(False)

        nodes = []

        nodes.append(tracker.root.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin)

        creator(origin)

        results = list(tracker.root.children)

        assert results[-2].node_type == "task"
        assert results[-2].text == "mid"
        assert results[:-2] + results[-1:] == nodes
        assert repr(creator)

    def test_before_explicit(self):
        creator = searching.parse_create_single(
                "<- task: first :{before, last}")
        assert creator.is_before
        assert creator.last_segment.tags == set()
        assert creator.last_segment.plurality == "last"

        tracker = Tracker(False)

        nodes = []

        nodes.append(tracker.root.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "started"))
        nodes[-1].start()
        nodes.append(tracker.root.createchild("task", "untouched"))
        nodes.append(tracker.root.createchild("task", "untouched 2"))

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin)

        creator(origin)

        results = list(tracker.root.children)

        assert results[0].node_type == "task"
        assert results[0].text == "first"
        assert results[1:] == nodes
        assert repr(creator)

    def test_after_match(self):
        creator = searching.parse_create_single("-> task: mid")
        assert creator.is_before
        assert creator.last_segment.tags == set(["can_activate"])
        assert creator.last_segment.plurality == "first"

        tracker = Tracker(False)

        nodes = []

        origin = tracker.root.createchild("task", "origin")
        nodes.append(origin)

        nodes.append(tracker.root.createchild("task", "finished"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "finished 2"))
        nodes[-1].start()
        nodes[-1].finish()
        nodes.append(tracker.root.createchild("task", "started 1"))
        nodes[-1].start()
        nodes.append(tracker.root.createchild("task", "untouched"))

        creator(origin)

        results = list(tracker.root.children)

        assert results[-3].node_type == "task"
        assert results[-3].text == "mid"
        assert results[:-3] + results[-2:] == nodes
        assert repr(creator)

    def test_other_tag(self):
        creator = searching.parse_create_single(
                "<- task: target :{first, started}")
        assert not creator.is_before
        assert creator.last_segment.tags == set(["started"])
        assert creator.last_segment.plurality == "first"
        assert repr(creator)

    def test_empty_prev_peer(self):
        tracker = Tracker(False)
        origin = tracker.root.createchild("task", "origin")
        creator = searching.parse_create("<- task: derp")
        creator(origin)
        assert match("\n".join(serialize(tracker.root, is_root=True)), (
            "task#?????: derp\n"
            "task#?????: origin"
        ))
        assert repr(creator)

    def test_empty_next_peer(self):
        tracker = Tracker(False)
        origin = tracker.root.createchild("task", "origin")
        creator = searching.parse_create("-> task: derp")
        creator(origin)
        assert match("\n".join(serialize(tracker.root, is_root=True)), (
            "task#?????: origin\n"
            "task#?????: derp"
        ))
        assert repr(creator)

    def test_default_first(self):
        creator = searching.parse_create("-> > task: target")

        tracker = Tracker(False, FakeNodeCreator(GenericActivate))

        origin = tracker.root.createchild("task", "origin")
        tracker.root.createchild("task", "expected")
        tracker.root.createchild("task", "ignore 1")
        tracker.root.createchild("task", "ignore 2")
        tracker.root.createchild("task", "ignore 3")

        creator(origin)

        assert match("\n".join(serialize(tracker.root, is_root=True)), (
            "task#?????: origin\n"
            "task#?????: expected\n"
            "    task#?????: target\n"
            "task#?????: ignore 1\n"
            "task#?????: ignore 2\n"
            "task#?????: ignore 3"
        ))
        assert repr(creator)

    def test_last_2(self):
        creator = searching.parse_create("-> :{last} > task: target")

        tracker = Tracker(False, FakeNodeCreator(GenericActivate))

        origin = tracker.root.createchild("task", "origin")
        tracker.root.createchild("task", "ignore 1")
        tracker.root.createchild("task", "ignore 2")
        tracker.root.createchild("task", "ignore 3")
        tracker.root.createchild("task", "expected")

        creator(origin)

        assert match("\n".join(serialize(tracker.root, is_root=True)), (
            "task#?????: origin\n"
            "task#?????: ignore 1\n"
            "task#?????: ignore 2\n"
            "task#?????: ignore 3\n"
            "task#?????: expected\n"
            "    task#?????: target"
        ))
        assert repr(creator)

    def test_create_invalid_day(self):
        creator = searching.parse_create_single("day: September 5, 2020")
        tracker = Tracker(False)

        with pytest.raises(LoadError):
            creator(tracker.root)

        assert tracker.root.find("day").first() is None

    #def test_many(self):
    #    creator = searching.parse_create("-> :{many} > task: TARGET")

    #    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    #    origin = tracker.root.createchild("task", "origin")
    #    tracker.root.createchild("task", "expected 1")
    #    tracker.root.createchild("task", "no longer expected 2")
    #    tracker.root.createchild("task", "no longer expected 3")

    #    creator(origin)

    #    assert match("\n".join(serialize(tracker.root, is_root=True)), (
    #        "task#?????: origin\n"
    #        "task#?????: expected 1\n"
    #        "    task#?????: TARGET\n"
    #        "task#?????: no longer expected 2\n"
    #        "task#?????: no longer expected 3\n"
    #    ))
    #    assert repr(creator)


# crashers
# :{first} < :{last} < :{last}
# :{first} < :{last} ->
# < :{last}
# derp :{ derp

def test_massive(makequery):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))
    for x in range(10):  # 10 nodes ....
        tracker.root.createchild("node")

    # this loop is to warm up the jit.
    for _ in range(5):
        initial = time.time()
        explosive = "><><><><><><><><><><><><><><><><><><>"
        # ... to the power of len([x for x in explosive if x == ">"]) ...
        massive_query = makequery(explosive)
        for x in range(3):
            with pytest.raises(searching.TooManyMatchesError):
                # ... should abort after a fixed number of ticks
                results = massive_query(tracker.root).list()
        final = time.time()
        delta = final - initial
        print(delta)

    # and said aborting of the search should be nice and early,
    # so the user doesn't see any performance issues
    if makequery is searching.parse_single:
        assert delta < 0.3
    else:
        assert delta < 0.6


def test_massive_ignore_overflow(makequery):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))
    for x in range(10):
        tracker.root.createchild("node")

    explosive = "><><><><><><><><><><><><><><><><><><>"
    massive_query = makequery(explosive)

    results = massive_query(tracker.root).ignore_overflow().list()
    assert len(results) > 10
    assert len(results) < searching.MAX_TICKS


def test_queries_structure():
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    query = searching.parse_single("> task: 'rigidly defined node'")
    queries = searching.Queries(query)

    origin = tracker.root.createchild("task", "origin")
    target1 = origin.createchild("task", "rigidly defined node")

    assert queries.queries == (query,)

    assert queries(origin).nodes().list() == [target1]
    assert queries(origin).list() == queries(origin).nodes().list()
    # assert list(queries(origin).results)
    # assert list(queries(origin).actions)
    assert queries(origin).first() is target1
    assert queries(origin).first() is queries(origin).one()

    with pytest.raises(searching.NoMatchesError):
        queries(target1).one()
    assert queries(target1).first() is None


def test_limit(makequery):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))
    for x in range(10):  # 10 nodes ....
        tracker.root.createchild("node")

    query = makequery("> *")

    assert len(query(tracker.root).list()) == 10
    assert len(query(tracker.root).limit(5).list()) == 5
    assert len(query(tracker.root).ignore_overflow().limit(5).list()) == 5


def test_parse_filters(monkeypatch):
    def _hook1(queries):
        queries.append(searching.parse_single("comment: derp derp"))
        return queries

    def _hook2(queries):
        queries.append(searching.parse_single("task: some random query"))
        return queries

    def _hook3(queries):
        return queries * 2

    monkeypatch.setattr(searching, "parsecreatefilters", [
        _hook1,
        _hook2
    ])
    monkeypatch.setattr(searching, "parseonlyfilters", [
        _hook3
    ])

    assert searching.parse("around the world") == searching.Queries(
            searching.parse_single("around the world"),
            searching.parse_single("comment: derp derp"),
            searching.parse_single("task: some random query"),

            searching.parse_single("around the world"),
            searching.parse_single("comment: derp derp"),
            searching.parse_single("task: some random query")
    )

    assert (searching.parse_create("comment: hoop doop") ==
        searching._Creators(creators=(
            searching.parse_create_single("comment: hoop doop"),
            searching.parse_create_single("comment: derp derp"),
            searching.parse_create_single("task: some random query")
        ))
    )


def test_create_errors(monkeypatch):
    with pytest.raises(searching.CantCreateError) as e:
        searching.parse_create("asdfsdgsdg")
    a = str(e).split("\n")
    assert len(a) == 1
    assert "full node" in a[0]

    with pytest.raises(searching.CantCreateError):
        searching.parse_create_single("sgdfgdf")

    with pytest.raises(searching.CantCreateError):
        searching.parse_create_single("< task: derp")

    with pytest.raises(searching.CantCreateError):
        searching.parse_create_single("*")

    def _hook3(queries):
        return queries * 2
    monkeypatch.setattr(searching, "parsecreatefilters", [
        _hook3
    ])

    with pytest.raises(searching.CantCreateError) as e:
        searching.parse_create("asdfsdgsdg")
    a = str(e.value).split("\n")
    assert len(a) >= 2
    assert "full node" in a[-1]
    assert "full node" in a[-2]


def test_not_found_creates(monkeypatch):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    def _hook2(queries):
        return [searching.parse_single("herp > task: derp")] + queries
    monkeypatch.setattr(searching, "parsecreatefilters", [
        _hook2
    ])

    creator = searching.parse_create("task: herpity derpity")

    node = creator(tracker.root)
    assert node.text == "herpity derpity"
    assert tracker.root.find("*").one() is node

    tracker.root.createchild("task", "herp")

    assert len(tracker.root.children) == 2
    node = creator(tracker.root)
    assert node.text == "derp"
    assert len(tracker.root.children) == 2
    assert tracker.root.find("herp > derp").one() is node


def test_failed_create(monkeypatch):
    tracker = Tracker(False, FakeNodeCreator(GenericActivate))

    def _hook2(queries):
        return [searching.parse_single("herp > task: derp")] + queries
    monkeypatch.setattr(searching, "parsecreatefilters", [
        _hook2
    ])

    creator = searching.parse_create("herk derk > task: herk")

    with pytest.raises(searching.NodeNotCreated):
        creator(tracker.root)


# to test: filters. create obeys filters. create only creates one node.
# create only uses first filter.



from __future__ import unicode_literals, print_function

if __name__ == "__main__":
    import sys
    from treeoflife import searching
    if "create" in sys.argv:
        dothing = searching.parse_create
    else:
        dothing = searching.parse

    todofile = "/Users/lahwran/.treeoflife/life"
    from treeoflife.tracker import Tracker
    import time
    tracker = Tracker()
    with open(todofile, "r") as reader:
        tracker.deserialize("file", reader)

    while True:
        querytext = raw_input("query: ")
        import subprocess
        subprocess.call(["clear"])
        print("query:", querytext)
        a = time.time()
        queryer = dothing(querytext)
        b = time.time()
        print(queryer, b - a)

        inittime = time.time()
        results = queryer(tracker.root)
        finishtime = time.time()

        print()

        if hasattr(results, "list"):
            results = results.list()
            for x in results[:1000]:
                print( " > ".join([
                    str(node) for node in x.iter_parents()][::-1]))
            print()
            print(len(results), finishtime - inittime)
        else:
            print(results)
            print()
            print(1, finishtime - inittime)



from __future__ import unicode_literals, print_function

import functools

import datetime
from treeoflife.parseutil import Grammar
from ometa.runtime import ParseError


from treeoflife.nodes.node import Option

date_format = "%B %d, %Y"
time_format = "%I:%M:%S %p"
datetime_format = "%s %s" % (date_format, time_format)


class TimeGrammar(Grammar):
    grammar = """
    days_name = ( 'd' ('ay' 's'? )? )
    hours_name = ( 'h' (('ou')? 'r' 's'?)? )
    minutes_name = ( 'm' ('in' ('ute' 's'?)?)? )
    seconds_name = ('s' ('ec' ('ond' 's'?)?)?)

    days_delta = ws number:days ws days_name ','? -> "days", days
    hours_delta = ws number:hours ws hours_name ','? -> "hours", hours
    minutes_delta = ws number:minutes ws minutes_name ','?
            -> "minutes", minutes
    seconds_delta = ws number:seconds ws seconds_name ','?
            -> "seconds", seconds
    timedelta = (
        (   days_delta |
            hours_delta |
            minutes_delta |
            seconds_delta
        )+:result ?(len(dict(result).keys()) == len(result))
        -> datetime.timedelta(**dict(result))
    )

    month = (
        ('jan' 'uary'? -> 1) |
        ('feb' ('r'? 'uary')? -> 2) |
        ('mar' 'ch'? -> 3) |
        ('apr' 'il'? -> 4) |
        ('may' -> 5) |
        ('jun' 'e'? -> 6) |
        ('jul' 'y'? -> 7) |
        ('aug' 'ust'? -> 8) |
        ('sep' ('t' 'ember'?)? -> 9) |
        ('oct' 'ober'? -> 10) |
        ('nov' 'ember'? -> 11) |
        ('dec' 'ember'? -> 12)
    )
    month_optional = month?:result anything* -> result

    date_primary = month:month wss number:day c_wss number:year
            -> year, month, day
    date_tomorrow = 'tomorrow'
            -> (datetime.datetime.now() + datetime.timedelta(days=1)).date()
    date_today = 'today' -> datetime.datetime.now().date()

    date = ((date_primary:d -> datetime.date(*d)) |
           date_tomorrow |
           date_today)

    # time
    time_peak = '12' ':' number_2dg:m (':' number_2dg)?:s (
            (c_wss 'am' -> (0, m, s or 0)) |
            (c_wss 'pm' -> (12, m, s or 0))
    )
    time_morning = number_2dg:h ':' number_2dg:m (':' number_2dg)?:s (
             (c_wss 'pm' ?(h < 12) -> (h + 12, m, s or 0)) |
             (c_wss 'am' ?(h < 12) -> (h, m, s or 0)) |
             ( -> (h, m, s or 0))
    )
    time = (time_peak | time_morning):t -> datetime.time(*t)

    datetime = date:d c_wss time:t -> datetime.datetime.combine(d, t)
    """
    bindings = {
        "datetime": datetime
    }


str_to_timedelta = TimeGrammar.wraprule("timedelta")
str_to_date = TimeGrammar.wraprule("date")
str_to_time = TimeGrammar.wraprule("time")
str_to_datetime = TimeGrammar.wraprule("datetime")


datetime_to_str = lambda dt: dt.strftime(datetime_format)
date_to_str = lambda date: date.strftime(date_format)
time_to_str = lambda time: time.strftime(time_format)


def timedelta_to_str(delta):
    days = delta.days
    hours = int(delta.seconds / 3600)
    minutes = int((delta.seconds / 60) - (hours * 60))
    seconds = int(delta.seconds - (hours * 3600) - (minutes * 60))
    result = []
    if days > 0:
        result.append("%dd" % days)
    if hours > 0:
        result.append("%dh" % hours)
    if minutes > 0:
        result.append("%dm" % minutes)
    if seconds > 0 or not len(result):
        result.append("%ds" % seconds)
    result_str = " ".join(result)
    return result_str

week_length = 7
month_length = 365.0 / 12.0
year_length = 365.25


def approx_delta(cur_date, other_date):
    if other_date == cur_date:
        return 'today'

    delta = other_date - cur_date
    days = abs(delta.days)
    if delta.days == 1:
        return 'tomorrow'
    elif delta.days == -1:
        return 'yesterday'
    elif days < week_length:
        value = days
        unit = 'day'
    elif days < month_length:
        value = float(days) / week_length
        unit = 'week'
    elif days < year_length:
        value = days / month_length
        unit = 'month'
    else:
        value = days / year_length
        unit = 'year'

    # plurals
    if value > 1:
        unit += 's'

    inted = value * 10
    derped = int(inted + 0.9999)
    herped = float(derped) / 10.0
    value = ("%.1f" % herped).rstrip("0").rstrip(".")

    if other_date < cur_date:
        return '%s %s ago' % (value, unit)
    else:
        return '%s %s' % (value, unit)


class DatetimeOption(Option):
    incoming = staticmethod(str_to_datetime)
    outgoing = staticmethod(datetime_to_str)


class TimedeltaOption(Option):
    incoming = staticmethod(str_to_timedelta)
    outgoing = staticmethod(timedelta_to_str)


class TimeOption(Option):
    incoming = staticmethod(str_to_time)
    outgoing = staticmethod(time_to_str)



from __future__ import unicode_literals, print_function

from treeoflife.tracker import Tracker


def test_archival():
    tracker = Tracker(False)
    tracker.deserialize("str",
        "archived: task#abcde: \xfctest\n"
        "    task#zxcvb: \xfcderp\n"
        "        task#qwert: \xfcderk\n"
        "    task#hjklo: \xfcherp\n"
    )

    assert tracker.serialize("str") == (
        "archived#abcde: task#abcde: \xfctest\n"
        "    @_af\n"
        "    archived#zxcvb: task#zxcvb: \xfcderp\n"
        "        @_af\n"
        "        archived#qwert: task#qwert: \xfcderk\n"
        "            @_af\n"
        "    archived#hjklo: task#hjklo: \xfcherp\n"
        "        @_af\n"
    )

    tracker.deserialize("str",
        "unarchive: archived#abcde: task#abcde: \xfctest\n"
        "    @_af\n"
        "    archived#zxcvb: task#zxcvb: \xfcderp\n"
        "        @_af\n"
        "        archived#qwert: task#qwert: \xfcderk\n"
        "            @_af\n"
        "    archived#hjklo: task#hjklo: \xfcherp\n"
        "        @_af\n"
    )

    assert tracker.serialize("str") == (
        "task#abcde: \xfctest\n"
        "    task#zxcvb: \xfcderp\n"
        "        task#qwert: \xfcderk\n"
        "    task#hjklo: \xfcherp\n"
    )



from __future__ import unicode_literals, print_function

import datetime

from treeoflife.nodes.node import Node, nodecreator
from treeoflife.nodes.tasks import BaseTask, ActiveMarker
from treeoflife.ordereddict import OrderedDict
from treeoflife.file_storage import parse_line
from treeoflife import timefmt


#######################
### generic nodes


@nodecreator("_gennode")
class GenericNode(Node):
    multiline = True

    def __init__(self, node_type="_gennode", text=None, parent=None,
            nodeid=None):
        self.metadata = OrderedDict()
        super(GenericNode, self).__init__(node_type, text, parent, nodeid)

    def setoption(self, option, value):
        self.metadata[option] = value

    def option_values(self, adapter=None):
        result = [(x, y, True) for x, y in self.metadata.items()]
        return result


@nodecreator("_genactive")
class GenericActivate(Node):
    options = (
        ActiveMarker(),
    )
    can_activate = True

    def __init__(self, *args, **kwargs):
        Node.__init__(self, *args, **kwargs)

    def start(self):
        pass

    def finish(self):
        pass

    def unfinish(self):
        pass


@nodecreator("archived")
class Archived(GenericNode):
    def __init__(self, node_type, text, *a, **kw):
        if kw.get("nodeid", None) is None and text:
            _, _, nodeid, _, _ = parse_line(text)
            kw["nodeid"] = nodeid
        GenericNode.__init__(self, node_type, text, *a, **kw)

    @classmethod
    def fromnode(cls, node, parent):
        if node.node_type == "archived":
            newnode = node.copy(parent=parent, children=False, nodeid=node.id)
        else:
            text = "%s#%s: %s" % (node.node_type, node.id, node.text)
            newnode = cls("archived", text, parent, nodeid=node.id)
            for option, value, show in node.option_values():
                if show:
                    newnode.setoption(option, value)
        newnode.setoption("_af", None)

        for child in list(node.children):
            child = child.detach()
            newnode.addchild(cls.fromnode(child, parent=newnode))
        return newnode

    def load_finished(self):
        if ("__af" in self.metadata or
                ("__af", None, True) in self.parent.option_values()):
            self.metadata["__af"] = None
            return

        prev_node = self.prev_neighbor
        next_node = self.next_neighbor
        parent = self.parent
        detached = self.detach()
        parent.addchild(self.fromnode(detached, parent=parent),
                before=next_node, after=prev_node)


@nodecreator("unarchive")
@nodecreator("unarchived")
class Unarchiver(GenericNode):
    def __init__(self, node_type, text, *a, **kw):
        if kw.get("nodeid", None) is None and text:
            _, _, nodeid, _, _ = parse_line(text)
            kw["nodeid"] = nodeid
        GenericNode.__init__(self, node_type, text, *a, **kw)

    @classmethod
    def unarchive(cls, node, parent):
        if node.node_type not in ("archived", "unarchive", "unarchived"):
            newnode = node.copy(parent=parent, children=False, nodeid=node.id)
        else:
            indent, ismetadata, nodeid, node_type, text = parse_line(node.text)
            assert not indent
            assert not ismetadata
            if node_type == "archived":
                indent, ismetadata, nodeid, node_type, text = parse_line(text)
                assert not indent
                assert not ismetadata

            newnode = node.root.nodecreator.create(node_type, text, parent,
                    nodeid=nodeid)
            for option, value, show in node.option_values():
                if show and option != "_af":
                    newnode.setoption(option, value)

        for child in list(node.children):
            child = child.detach()
            newnode.addchild(cls.unarchive(child, parent=newnode))

        return newnode

    def load_finished(self):
        prev_node = self.prev_neighbor
        next_node = self.next_neighbor
        parent = self.parent
        detached = self.detach()
        parent.addchild(self.unarchive(detached, parent=parent),
                after=prev_node, before=next_node)


#######################
### comments


@nodecreator("comment")
@nodecreator("answer")
@nodecreator("solution")
@nodecreator("IGNORE")
class Comment(Node):
    multiline = True

    options = (
        timefmt.DatetimeOption("time"),
    )

    def user_creation(self):
        self.time = datetime.datetime.now()


#######################
### todo


@nodecreator("todo")
class TodoItem(Node):
    children_of = ["todo bucket"]
    allowed_children = []
    multiline = True
    preferred_parent = "< :{last} > todo bucket"


@nodecreator("todo bucket")
class TodoBucket(Node):
    toplevel = True
    allowed_children = ["todo"]

    def load_finished(self):
        self.root.todo = self


class NoActiveMarker(ActiveMarker):
    def get(self, node):
        return False, None



from __future__ import unicode_literals, print_function

from weakref import WeakKeyDictionary
import logging

from treeoflife.nodes.node import Node, nodecreator, _NodeListRoot
from treeoflife.nodes.tasks import BaseTask
from treeoflife.util import setter

logger = logging.getLogger(__name__)


class ProxiedAttr(object):
    """
    Override the presence of an attribute on the proxy with a
    retrieval of some sort
    """
    def __init__(self, name, is_node=False, disable=None):
        self.name = name
        self.is_node = is_node
        self.disable = disable

    def __get__(self, instance, owning_class):
        if instance._px_target is None or (
                self.disable and getattr(instance, self.disable, False)):
            return getattr(instance, self.name + "_default", None)
        _my_name = self.name
        _is_node = self.is_node

        proxy_root = instance._px_root
        target_result = getattr(instance._px_target, self.name)
        assert not proxy_root.is_owned(target_result) or \
                proxy_root is target_result
        if self.is_node:
            return proxy_root.get_proxy(target_result)
        return target_result

    def __set__(self, instance, value):
        assert instance._px_target is not None
        assert self.is_node, ("Only node ProxiedAttrs can be written to; "
                "others should wrap inherited methods and class attrs")

        if instance._px_root.is_owned(value):
            value = value._px_target

        setattr(instance._px_target, self.name, value)


class ProxyNode(Node):
    """
               multiline: attr
                textless: attr
           text_required: attr
                toplevel: attr
            can_activate: attr
             children_of: attr
        allowed_children: attr
        preferred_parent: attr

                 options: AttributeError
            _option_dict: AttributeError !

                __init__: internal
          _init_children: unused
                children: node
              _next_node: node
              _prev_node: node

               __initing: auto
                    root: attr
               node_type: auto
                  parent: node
                    text: auto

               _validate: attr
           continue_text: attr

               setoption: attr
           option-values: attr

            iter_parents: inherit/indirect
      iter_flat_children: inherit/indirect
            iter_forward: inherit/indirect
           iter_backward: inherit/indirect
           next_neighbor: inherit/indirect
           prev_neighbor: inherit/indirect
                    find: inherit/indirect
                  create: inherit/indirect

                addchild: func/indirect
             createchild: inherit/indirect
             removechild: inherit/indirect
                  detach: inherit/indirect
                    copy: func/indirect

           make_skeleton: inherit/indirect
         children_export: func/stub

                   start: attr
                  finish: attr

           load_finished: inherit/stub
                auto_add: func/stub
            ui_dictify: attr
            search_texts: attr
             search_tags: attr
           user_creation: attr
            _search_tags: AttributeError
                 __str__: func/indirect
                _do_repr: func/indirect
                __repr__: inherit/indirect
    """
    multiline = ProxiedAttr("multiline")
    textless = ProxiedAttr("textless")
    text_required = ProxiedAttr("text_required")
    toplevel = ProxiedAttr("toplevel")
    can_activate = ProxiedAttr("can_activate")

    @property
    def options(self):
        raise AttributeError

    children_of = ProxiedAttr("children_of")
    allowed_children = ProxiedAttr("allowed_children")
    preferred_parent = ProxiedAttr("preferred_parent")

    parent = ProxiedAttr("parent", is_node=True)

    @property
    def root(self):
        return self._px_root.root

    _next_node = ProxiedAttr("_next_node", is_node=True)
    _prev_node = ProxiedAttr("_prev_node", is_node=True)
    children = ProxiedAttr("children", is_node=True, disable="_px_nochildren")

    setoption = ProxiedAttr("setoption")
    # option_values = ProxiedAttr("option_values")

    def option_values(self):
        # TODO: cheap hack
        result = self._px_target.option_values()
        realresult = []
        realresult.append(("active", None, self.root.active_node is self))
        for key, value, show in result:
            if key == "active":
                continue
            realresult.append((key, value, show))
        return realresult

    validate = ProxiedAttr("validate")
    continue_text = ProxiedAttr("continue_text")

    start = ProxiedAttr("start")
    finish = ProxiedAttr("finish")

    search_texts = ProxiedAttr("search_texts")
    search_tags = ProxiedAttr("search_tags")
    user_creation = ProxiedAttr("user_creation")

    _nonproxied = (
        "children_default",

        "_px_target",
        "_px_root",
        "_px_nochildren",
        "children",
        "referred_to",

        "iter_parents",
        "iter_flat_children",
        "iter_forward",
        "iter_backward",
        "next_neighbor",
        "prev_neighbor",
        "find",
        "create",
        "createchild",
        "removechild",
        "detach",

        "make_skeleton",
        "load_finished",
        "auto_add",
        "_do_repr",
        "id"
    )

    def __init__(self, proxy_root, target):
        if target is proxy_root:
            self._px_nochildren = True
            self.children_default = ReferenceNodeList(proxy_root, None)
            object.__setattr__(self, "text", "<recursing>")
        self._px_target = target
        self._px_root = proxy_root
        self.referred_to = set()

    def addchild(self, child, before=None, after=None):
        before = self._px_root.unwrap(before)
        after = self._px_root.unwrap(after)
        child = self._px_root.unwrap(child)

        if child.parent is not None:
            assert self._px_root.unwrap(child.parent) is self._px_target
            child.parent = None

        result = self._px_target.addchild(child, before=before, after=after)
        return self._px_root.get_proxy(result)

    def copy(self, parent=None, children=True, options=True):
        if parent is None:
            parent = self.parent

        parent = self._px_root.unwrap(parent)
        return self._px_target.copy(parent, children, options)

    def detach(self):
        return self._px_target.detach()

    def children_export(self):
        return []

    def ui_dictify(self):
        return None

    def __getattr__(self, name):
        """
        Fallback in case there isn't a ProxiedAttr
        Most attributes will be retrieved this way; for instance,
        node type and text.
        """
        if name == "options":
            raise AttributeError("proxied nodes do not have"
                " direct access to the options list; use set_option "
                "and option_values")
        return getattr(self._px_target, name)

    def __setattr__(self, name, value):
        if name in ProxyNode._nonproxied:
            object.__setattr__(self, name, value)
        else:
            setattr(self._px_target, name, value)

    # this adds all class-level defined thingies to the list of non-proxied
    # values
    _nonproxied += tuple(locals().keys())
    _nonproxied = set(_nonproxied)


class ReferenceNodeList(_NodeListRoot):
    _next_node = ProxiedAttr("_next_node", is_node=True)
    _prev_node = ProxiedAttr("_prev_node", is_node=True)

    def __init__(self, proxy_root, target):
        self._next_node_default = self
        self._prev_node_default = self

        self._px_root = proxy_root
        self._px_target = target

    @property
    def length(self):
        if self._px_target is None:
            return 0
        return self._px_target.length

    def insert(self, child, before=None, after=None):  # pragma: no cover
        raise Exception("this should never be called")

    def remove(self, child):
        child = self._px_root.unwrap(child)

        self._px_target.remove(child)


@nodecreator("reference")
@nodecreator("work on")
@nodecreator("workon")
class Reference(BaseTask):

    def __init__(self, *args, **kwargs):
        self.proxies = WeakKeyDictionary()
        self._px_root = self
        self._real_px_target = None
        self._px_didstart = False
        self._px_dostart = False
        self._propagate_started = True
        BaseTask.__init__(self, *args, **kwargs)

        #if self.node_type in ["reference", "work on"]:
        #    self.node_type = "reference"

    def addchild(self, child, before=None, after=None):
        before = self._px_root.unwrap(before)
        after = self._px_root.unwrap(after)
        child = self._px_root.unwrap(child)

        if child.parent is not None:
            assert self._px_root.unwrap(child.parent) is self._px_target
            child.parent = None

        result = self._px_target.addchild(child, before=before, after=after)
        return self._px_root.get_proxy(result)

    @property
    def active_id(self):
        if self._px_target is not None:
            return self._px_target.active_id
        return self.id

    @setter
    def _px_target(self, newvalue):
        assert self._px_can_target or newvalue is None
        self._real_px_target = newvalue
        if self.finished is None:
            self.finished = getattr(newvalue, "finished", self.finished)
        self.children = self.get_proxy(newvalue.children)

    @setter
    def started(self, value):
        self._realstarted = value
        if (self._propagate_started
                and self._px_target is not None
                and value
                and not getattr(self._px_target, "started", False)):
            try:
                self._px_target.started = self.started
            except AttributeError:
                pass

    def _init_children(self):
        self.children = ReferenceNodeList(self, None)
        self._next_node = None
        self._prev_node = None

    def _find_target(self):
        _px_target = self.find(self.text).one()
        assert _px_target is not None, "No target found for %r" % self
        return _px_target

    @property
    def _px_can_target(self):
        return not self.finished

    def load_finished(self, target=None):
        if self._px_can_target:
            if target is not None:
                self._px_target = target
            else:
                self._px_target = self._find_target()
            self.text = "#%s" % self._px_target.id

        if self._px_dostart:
            self._px_start()
        self.started = self.started

    def is_owned(self, node):
        return getattr(node, "_px_root", None) is self

    def get_proxy(self, target):
        if target is None:
            return None
        if target is self._px_target:
            return self
        try:
            return self.proxies[target]
        except KeyError:
            is_nodelist = hasattr(target, "insert") and hasattr(target,
                    "remove")
            is_node = hasattr(target, "node_type") and hasattr(target, "text")

            assert is_nodelist != is_node  # != is boolean XOR

            if is_nodelist:
                newnode = ReferenceNodeList(self, target)
            else:
                newnode = ProxyNode(self, target)

            self.proxies[target] = newnode
            return newnode

    def unwrap(self, node):
        if not self.is_owned(node):
            return node
        assert node._px_target is not None
        return node._px_target

    def children_export(self):
        return []

    def start(self):
        if self._px_target is not None:
            self._px_start()
        else:
            self._px_dostart = True

    def _px_start(self):
        try:
            self._propagate_started = False

            BaseTask.start(self)
            if (self._px_target.can_activate
                    and not getattr(self._px_target, "started", None)):
                self._px_target.start()
                self._px_didstart = True
            self._px_dostart = False
        finally:
            self._propagate_started = True

    def _finish_cleanup(self):
        self._real_px_target = None
        self.children = ReferenceNodeList(self, None)
        self.proxies = None

    def finish(self):
        BaseTask.finish(self)
        self._finish_cleanup()

    def unfinish(self):
        temp_target = self._find_target()
        assert temp_target is not None, "cannot unfinish if my target is gone!"

        result = temp_target.unfinish()
        if not result:
            return result

        assert BaseTask.unfinish(self), ("was unable to unfinish reference!"
                    " tree state is probably bad!")
        assert not self.finished
        self.proxies = WeakKeyDictionary()
        self._px_dostart = False

        self.load_finished(temp_target)
        return True

    def ui_dictify(self, result=None):
        if result is None:
            result = {}
        result["target"] = self._px_target.id if self._px_target else None
        return BaseTask.ui_dictify(self, result)

    def __str__(self):
        if self._px_target is None:
            return Node.__str__(self)
        result = "%s: %s" % (self.node_type, self._px_target.text)
        return result.partition("\n")[0]


@nodecreator("depends")
@nodecreator("depend")
@nodecreator("dep")
class Depends(Reference):

    def __init__(self, *a, **kw):
        self._propagate_finished = True
        self._px_dofinish = False
        self._px_didfinish = False
        Reference.__init__(self, *a, **kw)

        if self.node_type in ["depend", "dep"]:
            self.node_type = "depends"

    def load_finished(self, target=None):
        Reference.load_finished(self, target)
        if self._px_dofinish:
            self._px_finish()
        self.finished = self.finished

        if self.finished:
            self._finish_cleanup()

    @setter
    def finished(self, value):
        self._realfinished = value

        if (self._propagate_finished
                and self._px_target is not None
                and value
                and not getattr(self._px_target, "finished", False)):
            try:
                self._px_target.finished = value
            except AttributeError:
                pass

    @property
    def _px_can_target(self):
        return True

    def finish(self):
        if self._px_target is not None:
            self._px_finish()
        else:
            self._px_dofinish = True

    def _px_finish(self):
        try:
            self._propagate_finished = False

            if (self._px_target.can_activate
                    and not getattr(self._px_target, "finished", None)):
                self._px_target.finish()
                self._px_didfinish = True
            self._px_dofinish = False
            Reference.finish(self)
        finally:
            self._propagate_finished = True



from __future__ import unicode_literals, print_function

import weakref
import datetime
import logging
import uuid

from twisted.internet.defer import Deferred

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def lazyproperty(func):
    """
    Crazy mixin thing to compensate for lack of __init__ on mixins

    Use this to create attributes that are initialized on use,
    instead of at __init__
    """
    sentinel = object()
    f = "_lazyproperty_" + str(uuid.uuid4())

    def prop_get(self):
        result = getattr(self, f, sentinel)
        if result is sentinel:
            result = func()
            setattr(self, f, result)
        return result

    def prop_set(self, new):
        setattr(self, f, new)

    return property(prop_get, prop_set, doc=func.__doc__)


class Alarm(object):
    _alarms_logger = logging.getLogger(__name__ + ".Alarm")

    def __init__(self, parent, callback, delta=None, date=None):
        self.twisted_timer = None
        self.timer_deferred = None
        self.root = None
        self.ready_root = None
        self.called = False
        self.parent = parent

        self.callback_func = callback
        self._date = None
        if delta is not None:
            self.delta = delta

        if date is not None:
            self.date = date

        self._alarms_logger.debug("alarm %r created: %r, %r, %r, %r",
                self, parent, callback, delta, date)

    def _node_ready(self, root):
        self._alarms_logger.debug("ready: %r, %r", self, root)
        self.root = root

    delta = property()

    @delta.setter
    def delta(self, newdelta):
        date = datetime.datetime.now() + newdelta
        self._alarms_logger.debug("setting %r.delta to %r: %r", self, newdelta,
                date)
        self.date = date

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, newdate):
        self._alarms_logger.debug("setting %r.date to %r", self, newdate)
        self._date = newdate
        self.called = False
        self.update()

    @property
    def active(self):
        return (self._is_safe()
                and self.date is not None
                and not self.called)

    def _is_safe(self):
        return (self.ready_root is self.root
                and self.root is not None
                and self.root.tracker.root is self.root
                and self.parent._next_node is not None
                and self.parent._next_node._prev_node is self.parent)

    def update(self):
        self._alarms_logger.debug("%r.update()", self)
        if self.twisted_timer and self.twisted_timer.active():
            self._alarms_logger.debug("%r.update(): clearing previous timer",
                    self)
            self.twisted_timer.cancel()
            self.twisted_timer = None
            self.timer_deferred = None

        if not self.active:
            self._alarms_logger.debug("%r.update(): not active",
                    self)
            return

        self.parent.alarms.add(self)
        now = datetime.datetime.now()
        if self.date < now:
            self._alarms_logger.debug("%r.update(): callback %r in past, "
                    "firing", self, now - self.date)
#            timer, deferred = self.root.tracker._set_alarm(
#                    callback=self._callback,
#                    delta=datetime.timedelta())
            self._callback(None)
            return
        else:
            self._alarms_logger.debug("%r.update(): setting alarm",
                    self)
            timer, deferred = self.root.tracker._set_alarm(
                    callback=self._callback,
                    date=self.date)
        self.twisted_timer = timer
        self.timer_deferred = deferred

    def _go_live(self):
        self._alarms_logger.debug("%r._go_live(), root=%r",
                self, self.root)
        self.ready_root = self.root
        self.update()

    def super_repr(self, obj):
        return {
            "id": str(id(obj)),
            "repr": repr(obj),
            "str": str(obj),
            "type": repr(type(obj))
        }

    def warn(self, message):
        nn = self.parent._next_node
        data = {
            "ready_root": self.super_repr(self.ready_root),
            "root": self.super_repr(self.root),
            "tracker_root": self.super_repr(self.root.tracker.root),
            "date": self.super_repr(self.date),
            "called": self.super_repr(self.called),
            "parent": self.parent,
            "parent_next_node": nn,
            "parent_rev_link": nn._prev_node if nn is not None else None
        }
        self._alarms_logger.warn("%s: %s %r", self, message, data)

    def _callback(self, tracker):
        if not self.active:
            self.warn("timer fired but alarm was inactive!")
            return
        self._alarms_logger.debug("%s calling callback", self)
        self.called = True

        # allow deallocation - no need to hold the reference once it's called
        # the callback can re-activate it though, which will pass through
        # self.update which will re-add it
        self.parent.alarms.remove(self)

        self.callback_func()


class NodeMixin(object):
    _alarms_logger = logging.getLogger(__name__ + ".NodeMixin")
    alarms = lazyproperty(set)

    def _insertion_hook(self):
        if getattr(self.root, "_add_alarm", None) is None:
            self._alarms_logger.warn("%r node inserted, but %r root "
                    "does not support alarms", self, self.root)
            return
        else:
            self._alarms_logger.debug("%r insertion hook", self)

        for alarm in list(self.alarms):
            alarm._node_ready(root=self.root)
            self.root._add_alarm(alarm)

    def alarm(self, *args, **kwargs):
        result = Alarm(self, *args, **kwargs)
        self.alarms.add(result)
        possible_root = getattr(self, "root", None)
        if getattr(possible_root, "_add_alarm", None) is not None:
            self._alarms_logger.debug("%r creating alarm (%r, %r)",
                    self, args, kwargs)
            result._node_ready(root=self.root)
            self.root._add_alarm(result)
        elif possible_root is not None:
            self._alarms_logger.warn("%r node created an alarm (%r, %r), "
                    "but %r root does not support alarms",
                    self, args, kwargs, self.root)
        else:
            self._alarms_logger.debug("%r node created an alarm (%r, %r), "
                    "but root is not set",
                    self, args, kwargs)
        return result


class RootMixin(object):
    _alarms_logger = logging.getLogger(__name__ + ".RootMixin")

    _alarm_ready = lazyproperty(lambda: False)
    alarms = lazyproperty(weakref.WeakSet)

    def _tracker_support_available(self):
        return getattr(self.tracker, "_set_alarm", None) is not None

    def _alarm_hook(self):
        if self._tracker_support_available():
            self._alarms_logger.info("%r root load_finished", self)
            for alarm in list(self.alarms):
                alarm._go_live()
            self._alarm_ready = True
        else:
            self._alarms_logger.warn("%r root load_finished, but %r tracker "
                    "does not support alarms", self, self.tracker)

    def _add_alarm(self, alarm):
        self.alarms.add(alarm)
        if self._alarm_ready:
            if self._tracker_support_available():
                self._alarms_logger.debug("%r _add_alarm when live: %r",
                    self, alarm)
                alarm._go_live()
            else:
                self._alarms_logger.error("%r root already marked ready, but "
                    "tracker no longer supports alarms (%s)", self,
                    self.tracker)
        else:
            self._alarms_logger.debug("%r _add_alarm before ready: %r",
                    self, alarm)


class ProxyCache(object):
    """
    Proxy for timers that will automatically unregister when a root is unloaded
    """
    _alarms_logger = logging.getLogger(__name__ + ".ProxyCache")

    def __init__(self, root, tracker):
        """
        root: the root node this timerproxy is associated with. tracker.root
                may change, which would cause this timerproxy to free itself
        tracker: tracker.root will be compared with root each time, and if
                they are not identity, then the timerproxy will be freed
        """
        self.root_ref = weakref.ref(root)
        self.tracker_ref = weakref.ref(tracker)
        self.timers = weakref.WeakSet()
        self.init_time = datetime.datetime.now()

    @property
    def alive(self):
        root = self.root_ref()
        if root is None:
            return False

        tracker = self.tracker_ref()
        if tracker is None:
            return False

        if root is not tracker.root:
            return False

        return True

    def make_callback(self, deferred):
        self._alarms_logger.debug("making callback for deferred %s", deferred)
        return self._make_callback(weakref.ref(deferred))

    def _make_callback(self, ref):
        def proxy():
            tracker = self.tracker_ref()
            if not self.alive:
                self.die()
                return

            # won't EVER happen, but in case it does ...
            assert tracker is not None, "tracker went jesus"

            deferred = ref()

            if deferred is None:
                logger.warn("deferred fired, target didn't exist")
                return

            deferred.callback(tracker)
        return proxy

    def add_timer(self, timer):
        self._alarms_logger.debug("adding timer: %r", timer)
        self.timers.add(timer)

    def die(self):
        self._alarms_logger.info("Freeing Alarm Proxy Cache initialized at %r",
                self.init_time)
        for timer in list(self.timers):
            try:
                timer.cancel()
            except ValueError:
                continue


class TrackerMixin(object):
    _alarms_logger = logging.getLogger(__name__ + ".TrackerMixin")
    _alarm_proxy_caches = lazyproperty(weakref.WeakKeyDictionary)

    def _set_alarm(self, callback=None, delta=None, date=None):
        """
        Set an alarm.
        IMPORTANT: you *must* retain a reference to the returned deferred!
        failure to do so can cause the callback to not fire. this function
        is normally called by an Alarm instance, which also must be held
        onto to stay live.
        """
        self._alarms_logger.debug("setting alarm: %r, %r, %r", callback,
                delta, date)
        deferred = Deferred()
        if callback is not None:
            self._alarms_logger.debug("adding callback")
            deferred.addCallback(callback)

        if delta is None:
            current_time = datetime.datetime.now()
            if date <= current_time:
                raise ValueError("wat? %r %r" % (date, current_time))
            delta = date - current_time
            self._alarms_logger.debug("using date to calculate delta: %s",
                    delta)

        seconds = delta.total_seconds()
        self._alarms_logger.debug("setting alarm for %s seconds", seconds)

        # weakref hack, weakref hacks everywhere
        # if you're just reading this code, you can
        # pretend that these next few lines are just:
        # reactor.callLater(seconds, deferred.callback, self)
        # this is only relevant if you're worried about reloading of self.root
        proxy = self._alarm_proxy_caches.setdefault(self.root,
                    ProxyCache(self.root, self))
        callback = proxy.make_callback(deferred)
        timer = self._reactor.callLater(seconds, callback)
        proxy.add_timer(timer)
        # and we return you to your regularly scheduled sanity

        return timer, deferred



from __future__ import unicode_literals, print_function

from treeoflife.exceptions import LoadError
from treeoflife.util import HandlerDict
import string


loaders = HandlerDict()
serializers = HandlerDict()

# would making this a set speed it up?
nodeidchars = unicode(string.letters + string.digits)


def parse_line(line):
    assert type(line) == unicode

    parsing_indent = 0
    parsing_type = 1
    parsing_id = 2
    parsing_type_text_sep = 3
    parsing_text = 4

    parsing = 0
    indent = 0
    id = None
    is_metadata = False

    node_type = u""
    text = None

    if line.strip() == u"":
        return 0, False, None, u"", None

    for char in line:
        last_parsing = parsing
        if parsing == parsing_indent:
            if char == u" ":
                indent += 1
                continue
            else:
                if indent % 4 != 0:
                    raise LoadError("bad indentation")
                indent = indent / 4.0

                parsing += 1

        if char == u"\n":  # pragma: no cover
            continue

        if parsing == parsing_type:
            if last_parsing == parsing_indent:
                if char == "@":
                    is_metadata = True
                    continue
                elif char == u"-":
                    node_type = u"-"
                    parsing = parsing_type_text_sep
                    continue
            if char == u":":
                parsing = parsing_type_text_sep
                continue
            if char == u"#":
                parsing = parsing_id
                continue

            node_type += char
            continue

        if parsing == parsing_id:
            if id is None:
                id = u""
            if char == u":":
                parsing += 1
                continue
            elif char not in nodeidchars:
                raise LoadError("node id must be a-zA-Z0-9, got: %s%s" %
                        (id, char))
            id += char
            continue

        if parsing == parsing_type_text_sep:
            if char != u" ":
                raise LoadError("must be space separated")
            parsing += 1
            continue

        if text is None:
            text = u""

        text += char

    if id is not None and len(id) != 5:
        raise LoadError("node id must be exactly 5 long")
    if id is not None and is_metadata:
        raise LoadError("cannot have node IDs on metadata")

    return indent, is_metadata, id, node_type, text


@loaders.add("str")
def parse_string(string):
    assert type(string) == unicode
    return FileParser(string.split(u'\n'), decode=False)


@loaders.add("file")
class FileParser(object):
    def __init__(self, reader, decode=True):
        self.reader = reader
        self.error_context = None
        self.decode = decode

    def __iter__(self):
        return parse_file(self.reader, self.error_context, decode=self.decode)


def parse_file(reader, error_context, decode=True):
    reader = list(reader)
    for index, line in enumerate(reader):
        error_context.line = index
        if decode:
            line = line.decode("utf-8")
        if index == len(reader) - 1 and not line:
            continue
        yield parse_line(line)


def serialize(tree, is_root=False, one_line=False):
    lines = []
    if is_root:
        indent = u""
    else:
        indent = u" " * 4

        if tree.node_type == u"":
            lines.append(u"")
        elif tree.text:
            text_lines = tree.text.split("\n")
            lines.append(u"%s#%s: %s" % (tree.node_type,
                tree.id, text_lines[0]))
            for line in text_lines[1:]:
                lines.append(indent + "- %s" % line)
        else:
            lines.append(u"%s#%s" % (tree.node_type, tree.id))

    for name, value, show in tree.option_values():
        if not show:
            continue

        if value is None:
            lines.append(u"%s@%s" % (indent, name))
        else:
            lines.append(u"%s@%s: %s" % (indent, name, value))

    for child in tree.children_export():
        for line in serialize(child):
            lines.append(indent + line)
    return lines


@serializers.add("str")
def serialize_to_str(root, is_root=True):
    lines = serialize(root, is_root=is_root)
    return u'\n'.join(lines) + u"\n"


@serializers.add("file")
def serialize_to_file(root, writer, encode=True):
    result = serialize_to_str(root)
    if encode:
        result = result.encode("utf-8")
    writer.write(result)



from __future__ import unicode_literals, print_function

import itertools

import pytest

from treeoflife.nodes.misc import GenericNode, GenericActivate
from treeoflife.nodes.node import TreeRootNode
from treeoflife.file_storage import serialize_to_str
from treeoflife.test.util import FakeNodeCreator, match
from treeoflife import exceptions

from treeoflife.tracker import Tracker


class TestTracker(object):
    def test_load_basic(self, tmpdir):
        tracker_obj = Tracker(nodecreator=FakeNodeCreator(),
                skeleton=False)
        f = tmpdir.join('file')
        with f.open("wb") as w:
            w.write((
                "firstchild: \u2028args\n"
                "    secondchild: \u2028other args\n"
                "        @option: \u2028data\n"
                "        thirdchild: \u2028herp derp\n"
                "    fourthchild: \u2028ark dark\n"
            ).encode("utf-8"))
        tracker_obj.deserialize("file", f.open("rb"))

        root = tracker_obj.root
        root_child0 = root.children.next_neighbor
        assert root_child0.id in root.ids
        assert len(root_child0.id) == 5
        assert root_child0.text == "\u2028args"
        assert root_child0.node_type == "firstchild"

        root_child0_child0 = root_child0.children.next_neighbor
        assert root_child0_child0.id in root.ids
        assert len(root_child0_child0.id) == 5
        assert root_child0_child0.text == "\u2028other args"
        assert root_child0_child0.node_type == "secondchild"
        assert root_child0_child0.metadata["option"] == "\u2028data"

        root_child0_child0_child0 = root_child0_child0.children.next_neighbor
        assert root_child0_child0_child0.text == "\u2028herp derp"
        assert root_child0_child0_child0.node_type == "thirdchild"

        root_child0_child1 = root_child0_child0.next_neighbor
        assert root_child0_child1.text == "\u2028ark dark"
        assert root_child0_child1.node_type == "fourthchild"

    def test_load_bad_order(self):
        tracker_obj = Tracker(nodecreator=FakeNodeCreator(),
                skeleton=False)

        lines = (
            "firstchild: args\n"
            "    secondchild: other args\n"
            "    @option: this won't work here\n"
        )
        with pytest.raises(exceptions.LoadError):
            tracker_obj.deserialize("str", lines)

    def test_load_continued_text(self):
        tracker = Tracker(skeleton=False)
        lines = (
            "_gennode: derp\n"
            "    - herp\n"
            "    - derp\n"
        )
        tracker.deserialize("str", lines)
        assert tracker.root.children.next_neighbor.text == "derp\nherp\nderp"

    def test_save_basic(self, tmpdir):
        tracker = Tracker(nodecreator=FakeNodeCreator(),
                skeleton=False)

        node1 = GenericNode("node1", "\u2665node1_text", tracker.root)
        tracker.root.addchild(node1)
        node2 = GenericNode("node2", "\u2665node2_text", tracker.root)
        tracker.root.addchild(node2)
        node3 = GenericNode("node3", None, tracker.root)
        tracker.root.addchild(node3)

        node1_1 = GenericNode("node1_1", "\u2665node1_1_text", node1)
        node1_1.setoption("herp", "\u2665derp")
        node1.addchild(node1_1)

        node1_2 = GenericNode("node1_2", "\u2665node1_2_text", node1)
        node1_2.continue_text("\u2665herk derk")
        node1.addchild(node1_2)

        node2_1 = GenericNode("node2_1", "\u2665node2_1_text", node2)
        node2_1.setoption("hark", "\u2665dark")
        node2_1.continue_text('\u2665honk donk')
        node2.addchild(node2_1)

        f = tmpdir.join("file")
        target = f.open("wb")
        with target:
            tracker.serialize("file", target)
        assert match(f.read("rb"), (
            "node1#?????: \u2665node1_text\n"
            "    node1_1#?????: \u2665node1_1_text\n"
            "        @herp: \u2665derp\n"
            "    node1_2#?????: \u2665node1_2_text\n"
            "        - \u2665herk derk\n"
            "node2#?????: \u2665node2_text\n"
            "    node2_1#?????: \u2665node2_1_text\n"
            "        - \u2665honk donk\n"
            "        @hark: \u2665dark\n"
            "node3#?????\n"
        ).encode("utf-8"))

    def test_empty_line(self):
        tracker = Tracker(skeleton=False)
        tracker.deserialize("str",
                "\n"
                "task#12345: whatever\n"
                "   \n"
                "    task#abcde: whatever again\n"
                "\n"
                "    task#hijkl: some other thing\n"
                "\n"
                "\n"
        )
        assert tracker.serialize("str") == (
                "\n"
                "task#12345: whatever\n"
                "    \n"
                "    task#abcde: whatever again\n"
                "        \n"
                "    task#hijkl: some other thing\n"
                "        \n"
                "        \n"
        )

    def test_too_indented(self):
        tracker = Tracker(skeleton=False,
                nodecreator=FakeNodeCreator(GenericActivate))
        with pytest.raises(exceptions.LoadError):
            tracker.deserialize("str",
                "herp\n"
                "    derp\n"
                "            donk\n"
            )

    def test_unhandled_load_error(self):
        class ExcOnOptionNode(GenericNode):
            def setoption(self, option, value):
                raise Exception("test exception")

        tracker = Tracker(skeleton=False,
                nodecreator=FakeNodeCreator(ExcOnOptionNode))
        input_str = (
            "node1: node1\n"
            "node2: node2\n"
            "node3: node3\n"
            "    @someoption: boom\n"
        )
        try:
            tracker.deserialize("str", input_str)
        except exceptions.LoadError as e:
            result = str(e)
            assert "At line 4: UNHANDLED ERROR" in result
            assert "Exception: test exception" in result
        else:  # pragma: no cover
            assert False

    def test_roottype(self):
        class SubRootNode(TreeRootNode):

            target_variable = "present"

        tracker = Tracker(skeleton=False, roottype=SubRootNode)
        assert tracker.roottype is SubRootNode
        assert type(tracker.root) is SubRootNode

        tracker.deserialize("str", "")

        assert type(tracker.root) is SubRootNode



from __future__ import unicode_literals, print_function

import itertools
import functools
from collections import deque, namedtuple

# Note: a lot of classes here seem to "not be classes",
# in the terms of the "stop writing classes" talk. however,
# almost all of them are *used as data* by other code, hence
# why there are classes that look like they're used as functions.
from treeoflife import parseutil
from treeoflife.util import HandlerDict, HandlerList, memoize
from treeoflife import file_storage
from treeoflife.exceptions import LoadError


_Node = namedtuple('Node', 'rel, type, text')
_TaggedPattern = namedtuple('TaggedPattern', 'pat, tags')


class Queries(object):
    def __init__(self, *queries):
        self.queries = queries

    def __call__(self, node, counter=None):
        return QueriesResults(self.queries, node)

    def __eq__(self, other):
        return self.queries == getattr(other, "queries", None)

    def __repr__(self):
        joined = "\n".join(("    " + repr(q)) for q in self.queries)
        return "<queries\n%s\n>" % (joined,)


class Query(object):
    def __init__(self, *segments):
        self.segments = tuple(segments)

    def copy(self):
        return Query(*(segment.copy() for segment in self.segments))

    def __call__(self, node, counter=None):
        return QueryResults(self.segments, node, counter)

    def __repr__(self):
        return "<query %r>" % (self.segments,)

    def __eq__(self, other):
        return self.segments == getattr(other, "segments", None)

pluralities = set(["many", "first", "last"])


def split_tags(orig_tags):
    tags = set()
    plurality = None
    for tag in orig_tags:
        if tag in pluralities:
            if not plurality is None:
                assert False
            plurality = tag
        else:
            tags.add(tag)
    return tags, plurality

retrievers = HandlerDict()


def make_segment(separator, nodeid, pattern, tags):
    from treeoflife.nodes.node import nodecreator

    if separator == "default":
        if nodeid is None:
            separator = "children"
        else:
            separator = "self"

    tags, plurality = split_tags(tags)

    rel, type, text = pattern
    if text is None:
        if type == "*":
            matcher = None
        elif nodecreator.exists(type):
            matcher = Matcher(type, "*")
        else:
            matcher = Matcher("*", type)
    else:
        matcher = Matcher(type, text, rel=rel)

    return Segment(separator, pattern,
            matcher, tags, plurality, nodeid=nodeid)


class SearchGrammar(parseutil.Grammar):
    grammar = r"""
    ws = ' '+
    text = <(~separator ~tags_begin ~': ' anything)+>
    quote = ('"' | "'")
    endquote :q = quote:q2 ?(q == q2)
    node_text = ((ws? quote:q
                    (('\\' ('\\' | quote):z -> z) | ~endquote(q) anything)+:r
                    endquote(q) ws? -> ''.join(r))
                | <(~separator ~tags_begin anything)+>)
    node_text_star = ('*' node_text? -> '*') | ~'*' node_text
    text_star = ('*' text? -> '*' ) | ~'*' text

    createrel = '+' | '-' | -> 'default'
    node_base = ws? createrel:rel ~"#" (text_star:node
                (~tags_begin ws? ':' ws node_text_star)?:text
                -> Node(rel, node, text)
                | -> Node(rel, '*', None))

    tagged_node = node_base:pattern (tags | -> ()):tags -> pattern, tags
    matcher :sep :nodeid = tagged_node:p
            -> make_segment(sep, nodeid, p[0], p[1])

    tags_begin = ws? ':{'
    tags_end = ws? '}' | ~~separator
    tags_sep = ws? ','
    tag_text = <(~tags_end ~tags_sep anything)+>
    tag_texts = ws? tag_text:tag1 (tags_sep ws? tag_text)*:tags
                -> (tag1,) + tuple(tags)
    tags = tags_begin tag_texts:tags tags_end -> tags


    separator = (ws? '->' -> "next_peer")
                | (ws? '<-' -> "prev_peer")
                | (ws? '<<' -> "root")
                | (ws? ('>' ws?)? '**' (ws? '>')? -> "flatten")
                | (ws? '>' -> "children")
                | (ws? '<' -> "parents")

    nodeidchar = anything:x ?(x in nodeidchars) -> x
    nodeid = ws? '#' <nodeidchar{5}>

    query = (nodeid:n (~~separator | ~anything) -> n
                | -> None):nodeid
            (separator | -> 'default'):initial_sep
            matcher(initial_sep nodeid):initial
            (separator:sep matcher(sep None))*:following
            -> Query(initial, *following)

    """
    bindings = {
        "Node": _Node,
        "TaggedPattern": _TaggedPattern,
        "Query": Query,
        "make_segment": make_segment,
        "nodeidchars": file_storage.nodeidchars
    }

parse_single = memoize(SearchGrammar.wraprule("query"))


parsecreatefilters = HandlerList()
parseonlyfilters = HandlerList()


@memoize
def parse(string=None, query=None):
    assert (string is not None) != (query is not None)
    query = parse_single(string) if string is not None else query
    queries = [query.copy()]

    for filter_ in parsecreatefilters:
        queries = filter_(queries)
    for filter_ in parseonlyfilters:
        queries = filter_(queries)

    return Queries(*queries)


@memoize
def parse_create(string=None, query=None, do_auto_add=False):
    assert (string is not None) != (query is not None)
    query = parse_single(string) if string is not None else query
    queries = [query.copy()]

    for filter_ in parsecreatefilters:
        queries = filter_(queries)

    return _Creators(queries, do_auto_add=do_auto_add)


def parse_create_single(string=None, query=None, do_auto_add=False):
    assert (string is not None) != (query is not None)
    query = parse_single(string) if string is not None else query
    query = query.copy()
    return _Creator(query, do_auto_add)


class _Creators(object):
    def __init__(self, queries=None, creators=None, do_auto_add=False):
        if creators is not None:
            self.creators = list(creators)
            return

        self.creators = []
        errors = []
        for query in queries:
            try:
                creator = _Creator(query, do_auto_add=do_auto_add)
            except CantCreateError as e:
                errors.append(e)
            else:
                self.creators.append(creator)
        if not self.creators:
            if len(errors) == 1:
                raise errors[0]

            raise CantCreateError(
                "Can't create multi, do to all being errors:\n%s" % (
                    "\n".join(str(x) for x in errors),
                )
            )

    def __call__(self, basenode):
        assert basenode.node_type, "Please provide a single node"
        for creator in self.creators:
            node = creator(basenode)
            if node is not None:
                return node
        raise NodeNotCreated

    def __repr__(self):
        return "<creators\n%s\n>" % (
            "\n".join(("    " + repr(q)) for q in self.creators),
        )

    def __eq__(self, other):
        return self.creators == getattr(other, "creators", None)


class CantCreateError(Exception):
    pass


class _Creator(object):
    def __init__(self, query, do_auto_add=False):
        self.query = query.copy()
        segment = self.query.segments[-1]
        self.query.segments = self.query.segments[:-1]

        if segment.matcher is None:
            raise CantCreateError("cannot create node without full node")
        if not segment.matcher.is_rigid:
            raise CantCreateError("cannot create node without full node")
        if segment.separator == "parents":
            raise CantCreateError("cannot create parent node")

        rel = segment.matcher.create_relationship
        if rel == "default":
            new_tags = set(segment.tags)

            if "before" in new_tags:
                if "after" in new_tags:
                    assert False
                new_tags.remove("before")
                self.is_before = True
            elif "after" in new_tags:
                new_tags.remove("after")
                self.is_before = False
            else:
                self.is_before = (segment.separator != "prev_peer")
                if not new_tags:
                    new_tags.add("can_activate")

            if segment.plurality is not None:
                assert segment.plurality != "many"
                new_tags.add(segment.plurality)
            else:
                new_tags.add("first")

        else:
            if segment.tags:
                assert False, "conflict between rel shortcut and tags"
            if segment.plurality is not None:
                assert False, (
                    "conflict between rel shortcut and tags")
            is_last = (rel == "+")

            if is_last:
                new_tags = set(["last"])
                self.is_before = False
            else:
                new_tags = set(["first"])
                self.is_before = True

            if segment.separator == "prev_peer":
                # invert order for prev_peer
                self.is_before = not self.is_before

        self.node_type = segment.matcher.type
        self.text = segment.matcher.text

        if segment.separator not in ("children", "next_peer", "prev_peer"):
            assert False

        new_last = make_segment(segment.separator, None,
                ("default", "*", None), new_tags)
        self.last_segment = new_last

        # is this really the right place to put auto add?
        # probably not, but where does it belong then?
        self.do_auto_add = do_auto_add

    def __eq__(self, other):
        return (
                self.do_auto_add == getattr(other, "do_auto_add", None)
            and self.query == getattr(other, "query", None)
            and self.is_before == getattr(other, "is_before", None)
            and self.node_type == getattr(other, "node_type", None)
            and self.text == getattr(other, "text", None)
            and self.last_segment == getattr(other, "last_segment", None)
        )

    def __call__(self, basenode):
        assert basenode.node_type, "Please provide a single node"
        parentnode = self.query(basenode).first()
        if parentnode is None:
            return

        new_node = parentnode.root.nodecreator.create(
                        self.node_type, self.text, None, validate=False)
        existing_nodes = list(self.last_segment(parentnode))
        try:
            if self.do_auto_add:
                parent = new_node.auto_add(creator=parentnode,
                        root=parentnode.root)
                if parent:
                    return new_node

            if existing_nodes:
                assert len(existing_nodes) == 1
                node = existing_nodes[0]

                if self.is_before:
                    rel = {"before": node}
                else:
                    rel = {"after": node}

                new_node = node.parent.addchild(new_node, **rel)
            elif self.last_segment.separator == "children":
                new_node = parentnode.addchild(new_node)
            elif self.last_segment.separator == "prev_peer":
                new_node = parentnode.parent.addchild(new_node,
                        before=parentnode)
            else:
                new_node = parentnode.parent.addchild(new_node,
                        after=parentnode.parent.children.prev_neighbor)
            new_node._validate()
        except LoadError:
            new_node.detach()
            raise

        return new_node

    def __repr__(self):
        return "<creator %r -> %r>" % (self.query, self.last_segment)


class TickCounter(object):
    "Mutable thing"
    def __init__(self):
        self.ticks = 0

try:
    import __pypy__
    MAX_TICKS = 100000
except ImportError:
    MAX_TICKS = 3500


def tick(counter):
    try:
        counter.ticks += 1
    except AttributeError:
        assert counter is None
        return

    if counter.ticks > MAX_TICKS:
        raise TooManyMatchesError()


@retrievers.add()
def retriever_children(node, counter=None):
    for child in node.children:
        tick(counter)
        yield child


@retrievers.add()
def retriever_flatten(node, counter=None):
    for depth, subnode in node.iter_flat_children():
        tick(counter)
        yield subnode


@retrievers.add()
def retriever_next_peer(node, counter=None):
    for peer in node.iter_forward():
        tick(counter)
        yield peer


@retrievers.add()
def retriever_prev_peer(node, counter=None):
    for peer in node.iter_backward():
        tick(counter)
        yield peer


@retrievers.add()
def retriever_parents(node, counter=None):
    for parent in node.iter_parents():
        tick(counter)
        if parent is node:
            continue
        yield parent


@retrievers.add()
def retriever_root(node, counter=None):
    tick(counter)
    yield node.root


@retrievers.add()
def retriever_self(node, counter=None):
    return [node]


def getnodeid(node, nodeid, counter=None):
    tick(counter)
    return node.root.ids[nodeid]


def tag_filter(nodes, tags, counter=None):
    tags = set(tags)
    for node in nodes:
        tick(counter)
        node_tags = set(node._search_tags())
        if tags <= node_tags:
            yield node


class Matcher(object):
    def __init__(self, type, text, rel="default"):
        self.type = None if type == "*" else type
        self.text = None if text == "*" else text
        self.create_relationship = rel

        self.is_rigid = self.type is not None and self.text is not None
        if not self.is_rigid and rel != "default":
            assert False

    def __call__(self, nodes, counter=None):
        for node in nodes:
            tick(counter)
            node_types, texts = node.search_texts()
            texts_lower = []
            for text in texts:
                if text is None:
                    continue
                texts_lower.append(text.lower())

            if self.type is not None and self.type not in node_types:
                continue
            if self.text is not None and self.text.lower() not in texts_lower:
                continue
            yield node

    def __repr__(self):
        return "Matcher(%r, %r, rel=%r)" % (self.type, self.text,
                self.create_relationship)

    def copy(self):
        return Matcher(self.type, self.text, self.create_relationship)

    def __eq__(self, other):
        sentinel = object()
        return (
                self.type == getattr(other, "type", sentinel)
            and self.text == getattr(other, "text", sentinel)
        )


class Segment(object):
    def __init__(self, separator, pattern,
            matcher, tags, plurality, nodeid=None):
        self.separator = separator  # string (informational only)
        self.pattern = pattern  # instance of _Node (informational only)

        retriever = retrievers.handlers["retriever_" + separator]
        self.retriever = retriever  # function
        self.matcher = matcher  # instance of Matcher
        self.tags = tags  # set
        self.plurality = plurality  # string

        self.nodeid = nodeid  # string or None

    def __call__(self, node, counter=None):
        if self.nodeid is not None:
            try:
                node = getnodeid(node, self.nodeid, counter)
            except KeyError:
                # since this function is a generator,
                # simply returning gives us the result we want:
                # no nodes yielded. it'll still be an iterator.
                return

        nodes = self.retriever(node, counter)
        if self.matcher is not None:
            nodes = self.matcher(nodes, counter)

        if self.tags:
            nodes = tag_filter(nodes, self.tags, counter)

        if self.plurality == "last":
            node = None
            for node in nodes:
                pass
            if node is not None:
                yield node
        elif self.plurality == "first":
            for node in nodes:  # pragma: no branch
                yield node
                break
        else:
            for node in nodes:
                tick(counter)
                yield node

    def copy(self):
        return Segment(
            self.separator,
            self.pattern,
            self.matcher.copy() if self.matcher is not None else None,
            set(self.tags) if self.tags is not None else None,
            self.plurality,
            self.nodeid
        )

    def __repr__(self):
        return "Segment(%r%s%s%s)" % (
                self.separator,
                ", m=" + repr(self.matcher) if self.matcher else "",
                ", t=" + repr(self.tags) if self.tags else "",
                ", p=" + repr(self.plurality) if self.plurality else "")

    def __eq__(self, other):
        sentinel = object()
        return (
                self.retriever == getattr(other, "retriever", sentinel)
            and self.matcher == getattr(other, "matcher", sentinel)
            and self.tags == getattr(other, "tags", sentinel)
            and self.plurality == getattr(other, "plurality", sentinel)
            and self.nodeid == getattr(other, "nodeid", sentinel)
        )


class _FancyIterator(object):
    def __init__(self, iterator):
        self.iterator = iterator

    def __iter__(self):
        return iter(self.iterator)

    def one(self):
        for x in self:
            return x
        raise NoMatchesError()

    def first(self):
        for x in self:
            return x
        return None

    def list(self):
        return list(self)

    def limit(self, length):
        return _FancyIterator(itertools.islice(self, length))

    def ignore_overflow(self):
        return _FancyIterator(_ignore_overflow(self.__iter__()))


def _ignore_overflow(iterator):
    try:
        for x in iterator:
            yield x
    except TooManyMatchesError:
        return


def fancify(func):
    @functools.wraps(func)
    def wrapper(*a, **kw):
        iterator = func(*a, **kw)
        return _FancyIterator(iterator)
    return wrapper


class _Frame(object):
    def __init__(self, node, iterator):
        self.node = node
        self.iterator = iter(iterator)
        self.found = False


class QueryResults(_FancyIterator):
    def __init__(self, segments, basenode, counter=None):
        assert basenode.node_type, "Please provide a single node"
        self.segments = segments
        self.basenode = basenode

        if counter is None:
            counter = TickCounter()
        self.counter = counter

    @fancify
    def nodes(self):
        return (x[2] for x in self._search(True, False))

    def __iter__(self):
        return iter(self.nodes())

    def _search(self, successes, errors):
        stack = deque()
        node = self.basenode
        stack.append(_Frame(node, [node]))
        # +1 for the initial frame
        seglen = len(self.segments)

        while stack:
            frame = stack[-1]
            position = len(stack) - 1

            try:
                node = frame.iterator.next()
            except StopIteration:
                if not frame.found and errors:
                    yield False, position, frame.node
                stack.pop()
                continue
            frame.found = True

            if position >= seglen:
                if successes:
                    yield True, None, node
                continue

            n_seg = self.segments[position]
            stack.append(_Frame(node, n_seg(node, self.counter)))


class QueriesResults(_FancyIterator):
    def __init__(self, queries, basenode, counter=None):
        assert basenode.node_type, "Please provide a single node"

        self.basenode = basenode

        if counter is None:
            counter = TickCounter()
        self.counter = counter

        self.queries = []
        for query in queries:
            self.queries.append(query(basenode, counter))

    def _makechain(self, func):
        return itertools.chain.from_iterable(
                func(query) for query in self.queries)

    @fancify
    def nodes(self):
        return self._makechain(lambda q: q.nodes())

    def _search(self, successes, errors):
        return self._makechain(lambda q: q._search(successes, errors))

    def __iter__(self):
        return iter(self.nodes())


def one(iterator):
    return _FancyIterator(iterator).one()


def chain(*queries):
    def run(basenode, counter=None):
        assert basenode.node_type, "Please provide a single node"
        if counter is None:
            counter = TickCounter()
        for query in queries:
            for result in query(basenode, counter):
                yield result
    return run


class NoMatchesError(Exception):
    pass


class NodeNotCreated(NoMatchesError):
    pass


class TooManyMatchesError(Exception):
    pass



from __future__ import unicode_literals, print_function

from treeoflife.userinterface import command
from treeoflife import searching


def _create(query, root, auto=False):
    creator = searching.parse_create(query=query, do_auto_add=auto)
    node = creator(root.active_node)
    node.user_creation()

    return node


def _filter_can_activate(nodes):
    # TODO: this is a dumb solution
    for node in nodes:
        good = True
        for parent in node.iter_parents():
            if parent.node_type in ["days", "life"]:
                break
            if not parent.can_activate:
                good = False
                break
        if good:
            yield node


def _activate(nodes, root, force=False):
    if not force:
        # this is a hackfix for the real problem, but will do for a while
        nodes = _filter_can_activate(nodes)
    node = searching.one(nodes)

    active = root.active_node
    root.activate(node, force=force)
    return active


@command()
@command("next")
def done(root):
    query = searching.parse_single("> * :{can_activate}")
    to_activate = query(root.active_node).first()

    if to_activate is not None:
        _activate([to_activate], root)
        return

    query = searching.chain(
        searching.parse_single("-> :{can_activate}"),
        searching.parse_single("< :{can_activate}")
    )
    nodes = query(root.active_node)

    try:
        prev_active = _activate(nodes, root)
    except searching.NoMatchesError:
        return
    else:
        prev_active.finish()


@command()
def createauto(text, root):
    query = searching.parse_single(text)
    try:
        activate(text, root)
        return
    except searching.NoMatchesError:
        message = "no existing node to activate"

    try:
        createactivate(text, root)
        return
    except searching.NodeNotCreated:
        message += ", and parent node to create under not found"
    except searching.CantCreateError as e:
        message += "; " + str(e)
    raise Exception(message)


@command()
@command("c")
def create(text, root):
    query = searching.parse_single(text)
    return _create(query, root)


@command()
@command("a")
def activate(text, root):
    query = searching.parse(text)
    nodes = query(root.active_node)
    return _activate(nodes, root)


@command()
@command("fa")
def forceactivate(text, root):
    query = searching.parse(text)
    nodes = query(root.active_node)
    return _activate(nodes, root, force=True)


@command()
@command("f")
def finish(text, root):
    query = searching.parse(text)
    nodes = query(root.active_node)
    prev_active = _activate(nodes, root)
    prev_active.finish()


@command("ca")
def createactivate(text, root):
    query = searching.parse_single(text)
    node = _create(query, root, auto=True)

    try:
        prev = _activate([node], root)
    except searching.NoMatchesError:
        # couldn't find activate tag on node. create, but do nothing
        return

    return prev


@command("cf")
def createfinish(text, root):
    prev_active = createactivate(text, root)
    if prev_active is not None:
        prev_active.finish()



from __future__ import unicode_literals, print_function

import traceback
import subprocess
import os
import logging
import uuid
import json

from treeoflife.userinterface import command
from treeoflife.util import tempfile, Profile, HandlerDict

logger = logging.getLogger(__name__)


editor_types = HandlerDict()


@command()
def editpudb(ui, source):  # pragma: no cover
    import pudb
    pudb.set_trace()
    edit(ui, source)


@command()
def edit(ui, source):
    ui.edit_session = EditSession(ui, source)


class EditSession(object):
    def __init__(self, ui, source):
        self.ui = ui
        self.source = source
        name = ui.run_config.editor
        self.EditorType = editor_types.handlers[name]

        if self.EditorType.hide:
            ui.hide_all_clients()

        self.exceptions = []
        self.original_text = ui.serialize("str")
        self.old_root = self.ui.root

        self.editor = self.EditorType(self, self.original_text)

    def editor_done(self, edited_text):
        if edited_text == self.original_text:
            logger.info("text same, not loading")
            self.succeeded()
            return True

        try:
            with Profile("deserialize"):
                self.ui.deserialize("str", edited_text)
        except Exception as e:
            logger.exception("Failure loading")
            self.source.capture_error(e)
            formatted = traceback.format_exc()
            self.ui.root = self.old_root
            self.editor = self.EditorType(self, edited_text, error=formatted)
        else:
            logger.info("loaded; new active: %r", self.ui.root.active_node)
            self.succeeded()
            return True
        return False

    def succeeded(self):
        self.editor = None
        self.ui.edit_session = None
        self.ui.update_all()
        if self.EditorType.hide:
            self.ui.show_client(self.source)


class _TerminalLauncher(object):
    hide = True
    outer_command = b'exec bash -c "%s"\n'

    base_inner_command = (b"vim %s;echo '%s' "
                b"| base64 --decode| nc 127.0.0.1 %d")

    def __init__(self, session, contents, error=None):
        self.session = session
        self.filenames = []
        self.tmp = tempfile()
        self.tmp_backup = tempfile()
        self.filenames.append(self.tmp)
        port = session.ui.run_config.port

        data = contents.encode("utf-8")
        with open(self.tmp, "w") as writer:
            writer.write(data)
        with open(self.tmp_backup, "w") as writer:
            writer.write(data)

        self.error = error
        if error is not None:
            self.error_tmp = tempfile()
            with open(self.error_tmp, "w") as writer:
                writer.write(error.encode("utf-8"))
            self.filenames.append(self.error_tmp)

        args = self.shell_escape([b"-o", b"--"] + list(self.filenames))

        self.identifier = str(uuid.uuid4())
        self.json_data = json.dumps({
            "file_editor_finished": self.identifier}) + "\n"
        b64_data = self.json_data.encode("base64")
        b64_data = b64_data.replace(b" ", b"").replace(b"\n", b"")

        inner_command = self.base_inner_command % (
                b" ".join(args), b64_data, port)
        inner_command = inner_command.replace(b"\\'", b"'\"'\"'")
        self.command = self.outer_command % inner_command
        logger.info("Starting terminal editor with id %r: %s",
                self.identifier, self.command)

    def shell_escape(self, args):
        # Yes, I'm trying to do shell escaping. restricted inputs, so I'm not
        # too concerned; user input will never be passed to this.
        wrapped_args = []
        for arg in args:
            assert "'" not in arg, ("Can't put quotes in args! (sorry, "
                                "I know it's awkward and hacky)")
            wrapped_args.append(b"'%s'" % arg)
        return wrapped_args

    def done(self):
        with open(self.tmp, "r") as reader:
            data = reader.read().decode("utf-8")
        self.session.editor_done(data)

    def command_attempted(self):
        pass

    def source_died(self):
        pass


@editor_types.add("vim-iterm")
class ItermLauncher(_TerminalLauncher):  # pragma: no cover
    """
    Tell iterm2 to open a new window, then run a command that runs $EDITOR;
    after $EDITOR finishes, the command will send a json message to the main
    port
    """

    applescript = b"""
        tell application "iTerm"
            activate
            set myterm to (make new terminal)
            tell myterm
                set number of columns to 140
                set number of rows to 150

                launch session "Default Session"

                tell the last session
                    write contents of file "{tempfile}"
                end tell
            end tell
        end tell
    """

    def __init__(self, *a, **kw):
        _TerminalLauncher.__init__(self, *a, **kw)

        self.tempfile = tempfile()
        with open(self.tempfile, "w") as temp_writer:
            temp_writer.write(self.command)

        self.osascript(self.applescript.format(tempfile=self.tempfile))

    def command_attempted(self):
        self.osascript(
            b'tell application "iTerm"\n'
            b'    activate\n'
            b'end tell\n'
        )

    def osascript(self, code):
        temp_code = tempfile()
        with open(temp_code, "w") as code_writer:
            code_writer.write(code)
        subprocess.call([b"osascript", temp_code])
        os.unlink(temp_code)


@editor_types.add("embedded")
class EmbeddedEditor(object):
    hide = False

    def __init__(self, session, contents, error=None):
        self.session = session
        self.identifier = str(uuid.uuid4())

        self.message = {
            "embedded_edit": {
                "identifier": self.identifier,
                "data": contents,
                "error": error
            }
        }
        self.logfile = tempfile()
        with open(self.logfile, "a") as writer:
            writer.write(json.dumps(self.message))
            writer.write('\n')
        logger.info("Starting embedded editor (logfile %s)",
                self.logfile)
        self.session.source.sendmessage(self.message)

    def done(self, data):
        with open(self.logfile, "a") as writer:
            writer.write(json.dumps(data))
            writer.write('\n')
        if data is None:
            data = self.session.original_text
        logger.info("attempting to stop embedded editor")
        if self.session.editor_done(data):
            self.session.source.sendmessage({"embedded_edit": None})

    def command_attempted(self):
        pass

    def source_died(self):
        self.session.succeeded()



from __future__ import unicode_literals, print_function

from .node import Node, TreeRootNode, Option
from .tasks import BaseTask, Task, ActiveMarker
from .misc import (Comment, GenericNode, TodoItem,
        TodoBucket)
from .recordkeeping import FitnessLog, Weight, Calories, Waist, Workout
import treeoflife.nodes.references
from .days import Day, Days

import treeoflife.nodes.testnodes



from __future__ import unicode_literals, print_function

import datetime

from ometa.runtime import ParseError
import pytest

from treeoflife import timefmt


def test_string_date():
    assert timefmt.str_to_date("june 7, 2012") == datetime.date(2012, 6, 7)


def test_string_tomorrow(setdt):
    setdt(2012, 6, 7)
    assert timefmt.str_to_date("tomorrow") == datetime.date(2012, 6, 8)


def test_string_today(setdt):
    setdt(2012, 6, 7)
    assert timefmt.str_to_date("today") == datetime.date(2012, 6, 7)


def test_string_time():
    assert timefmt.str_to_time("10:00 AM") == datetime.time(10)


def test_string_datetime():
    target = datetime.datetime(2012, 6, 7, 10)
    assert timefmt.str_to_datetime("june 7, 2012 10:00 AM") == target


def test_time_string():
    assert timefmt.time_to_str(datetime.time(10)) == "10:00:00 AM"


def test_timedelta_to_str():
    x = datetime.timedelta(days=2, hours=2, minutes=2, seconds=2)
    assert timefmt.str_to_timedelta(timefmt.timedelta_to_str(x)) == x


class TestParsers(object):
    def test_months(self):
        with pytest.raises(ParseError):
            timefmt.TimeGrammar("Marceh"   ).month()
        with pytest.raises(ParseError):
            timefmt.TimeGrammar("Apiril"   ).month()
        assert timefmt.TimeGrammar("Apiril").month_optional() is None
        with pytest.raises(ParseError):
            timefmt.TimeGrammar("janruary").month()
        with pytest.raises(ParseError):
            timefmt.TimeGrammar("ferbuary").month()

        assert timefmt.TimeGrammar("jan"      ).month() == 1
        assert timefmt.TimeGrammar("january"  ).month() == 1
        assert timefmt.TimeGrammar("January"  ).month() == 1
        assert timefmt.TimeGrammar("feb"      ).month() == 2
        assert timefmt.TimeGrammar("febuary"  ).month() == 2  # misspelling
        assert timefmt.TimeGrammar("february" ).month() == 2
        assert timefmt.TimeGrammar("February" ).month() == 2
        assert timefmt.TimeGrammar("mar"      ).month() == 3
        assert timefmt.TimeGrammar("march"    ).month() == 3
        assert timefmt.TimeGrammar("March"    ).month() == 3
        assert timefmt.TimeGrammar("apr"      ).month_optional() == 4
        assert timefmt.TimeGrammar("april"    ).month() == 4
        assert timefmt.TimeGrammar("apr"      ).month() == 4
        assert timefmt.TimeGrammar("may"      ).month() == 5
        assert timefmt.TimeGrammar("jun"      ).month() == 6
        assert timefmt.TimeGrammar("june"     ).month() == 6
        assert timefmt.TimeGrammar("jul"      ).month() == 7
        assert timefmt.TimeGrammar("july"     ).month() == 7
        assert timefmt.TimeGrammar("august"   ).month() == 8
        assert timefmt.TimeGrammar("aug"      ).month() == 8
        assert timefmt.TimeGrammar("sep"      ).month() == 9
        assert timefmt.TimeGrammar("sept"     ).month() == 9
        assert timefmt.TimeGrammar("september").month() == 9
        assert timefmt.TimeGrammar("oct"      ).month() == 10
        assert timefmt.TimeGrammar("october"  ).month() == 10
        assert timefmt.TimeGrammar("nov"      ).month() == 11
        assert timefmt.TimeGrammar("november" ).month() == 11
        assert timefmt.TimeGrammar("dec"      ).month() == 12
        assert timefmt.TimeGrammar("december" ).month() == 12

    def test_time(self):
        assert timefmt.TimeGrammar("12:00 am").time() == datetime.time(0, 0)
        assert timefmt.TimeGrammar("00:00 am").time() == datetime.time(0, 0)
        assert timefmt.TimeGrammar("00:00"   ).time() == datetime.time(0, 0)
        assert timefmt.TimeGrammar("1:00 am" ).time() == datetime.time(1, 0)
        assert timefmt.TimeGrammar("1:00"    ).time() == datetime.time(1, 0)
        assert timefmt.TimeGrammar("01:00"   ).time() == datetime.time(1, 0)
        assert timefmt.TimeGrammar("09:00"   ).time() == datetime.time(9, 0)

        with pytest.raises(ParseError):
            timefmt.TimeGrammar("13:00 am").time()
        with pytest.raises(ParseError):
            timefmt.TimeGrammar("13:00 pm").time()
        with pytest.raises(ParseError):
            timefmt.TimeGrammar("1300").time()
        with pytest.raises(ParseError):
            timefmt.TimeGrammar("1300").time()

        assert timefmt.TimeGrammar("12:00"   ).time() == datetime.time(12, 0)
        assert timefmt.TimeGrammar("12:00 pm").time() == datetime.time(12, 0)
        assert timefmt.TimeGrammar("00:00 pm").time() == datetime.time(12, 0)
        assert timefmt.TimeGrammar("01:00 pm").time() == datetime.time(13, 0)
        assert timefmt.TimeGrammar("13:00"   ).time() == datetime.time(13, 0)


def test_approx_delta():
    now = datetime.date(2012, 12, 26)
    approx_delta = timefmt.approx_delta
    assert approx_delta(now, datetime.date(2012, 12, 31)) == '5 days'
    assert approx_delta(now, datetime.date(2013, 12, 31)) == '1.1 years'
    assert approx_delta(now, datetime.date(2013, 1, 31)) == '1.2 months'
    assert approx_delta(now, datetime.date(2013, 4, 7)) == '3.4 months'
    assert approx_delta(now, datetime.date(2013, 3, 21)) == '2.8 months'
    assert approx_delta(now, datetime.date(2013, 1, 15)) == '2.9 weeks'
    assert approx_delta(now, datetime.date(2012, 12, 27)) == 'tomorrow'
    assert approx_delta(now, datetime.date(2012, 12, 26)) == 'today'
    assert approx_delta(now, datetime.date(2012, 12, 25)) == 'yesterday'
    assert approx_delta(now, datetime.date(2011, 12, 25)) == '1.1 years ago'
    assert approx_delta(now, datetime.date(2011, 12, 26)) == '1.1 years ago'
    # can't get 1 year because 365.25
    assert approx_delta(now, datetime.date(2011, 12, 27)) == '12 months ago'
    assert approx_delta(now, datetime.date(2012, 12, 20)) == '6 days ago'
    assert approx_delta(now, datetime.date(2012, 12, 19)) == '1 week ago'



from __future__ import unicode_literals, print_function

import platform
import traceback
import os
import json
import subprocess
from functools import partial
import datetime
import time
import logging
import inspect
import itertools
import glob

from treeoflife.file_storage import parse_line
from treeoflife.nodes.node import nodecreator, TreeRootNode
from treeoflife.tracker import Tracker
from treeoflife.exceptions import InvalidInputError
from treeoflife.util import tempfile, HandlerDict, Profile
from treeoflife.parseutil import Grammar
from treeoflife import timefmt
from treeoflife import alarms

logger = logging.getLogger(__name__)


def _makenode(string):
    indent, is_metadata, nodeid, node_type, text = parse_line(string)
    if is_metadata:
        raise InvalidInputError("metadata not allowed")
    if indent > 0:
        raise InvalidInputError("plz 2 not indent")
    return node_type, text


global_commands = HandlerDict()
command = global_commands.add


import treeoflife.navigation


def _listing_node(active, node, indent):
    indent_text = " " * 4
    if active is node:
        prefix = "> "
    else:
        prefix = "  "
    return prefix + (indent_text * indent) + str(node)


def generate_listing(active, node, lines=None, indent=0):
    if lines is None:
        lines = []

    lines.append(_listing_node(active, node, indent))

    for child in node.children:
        generate_listing(active, child, lines, indent + 1)
    return lines


class Event(object):
    def __init__(self, source, root, command_name, text, ui):
        self.source = source
        self.root = root
        self.command_name = command_name
        self.text = text
        self.ui = ui
        self.event = self

    def _inject(self, function):
        func_args = inspect.getargs(function.__code__).args
        call = {}
        for arg in func_args:
            call[arg] = getattr(self, arg)
        return function(**call)


class MixedAlarmRoot(TreeRootNode, alarms.RootMixin):
    pass


class CommandInterface(Tracker, alarms.TrackerMixin):
    max_format_depth = 2
    _default_command = "createauto"

    def __init__(self, *args, **kwargs):
        kwargs["roottype"] = MixedAlarmRoot
        self._reactor = kwargs.pop("reactor")
        Tracker.__init__(self, *args, **kwargs)

    def _command(self, source, command_name, text):
        logger.info("command begin: %r, %r, %r", source, command_name, text)
        initial = time.time()
        try:
            target = global_commands.handlers[command_name]
        except KeyError:
            self.errormessage(source, "no such command %r" % command_name)
        else:
            event = Event(source, self.root, command_name, text, self)
            event._inject(target)
            final = time.time()
            logger.info("command took: %r", final - initial)

    def command(self, source, line):
        if not line.strip():
            return

        command_name, center, command_text = line.partition(" ")
        if command_name not in global_commands.handlers:
            command_text = line
            command_name = self._default_command

        self._command(source, command_name, command_text)

    def errormessage(self, source, message):
        logger.error(message)

    def display_lines(self, lines):
        pass

    def displaychain(self, limit=True):
        if limit:
            return [x[0] for x in zip(self.root.active_node.iter_parents(),
                                      range(self.max_format_depth))]
        else:
            return list(self.root.active_node.iter_parents())

    def tree_context(self, max_lines=55):
        active = self.root.active_node
        current = active.find("<day").one()
        days = current.parent
        root = self.root

        lines = []
        lines.append(_listing_node(active, days, 0))
        if current.prev_neighbor:
            lines.append(_listing_node(None, "...", 1))
        while len(lines) < max_lines and current:
            generate_listing(active, current, lines, indent=1)
            current = current.next_neighbor
        return lines


class Git(object):
    def __init__(self, path):
        self.path = path

        self.binary = self._which("git")

    def _which(self, name):
        """
        Borrowed from twisted.python.procutils, but modified so I can edit PATH
        """
        flags = os.X_OK
        result = []
        exts = filter(None, os.environ.get('PATHEXT', '').split(os.pathsep))
        path = [x for x in os.environ.get('PATH', '').split(os.pathsep) if x]

        if platform.system() == "Windows":
            directories = glob.glob("c:/*rogram*iles*/*git*/*bin*/")
            path.extend(directories)

        if not path:
            return []
        for p in path:
            p = os.path.join(p, name)
            if os.access(p, flags):
                result.append(p)
            for e in exts:
                pext = p + e
                if os.access(pext, flags):
                    result.append(pext)

        if not result:
            paths = "\n".join("    %s" % x for x in path)
            raise RuntimeError("Couldn't find %s in:\n%s" % (name, paths))

        return result[0]

    def init(self):
        if not os.path.isdir(os.path.join(self.path, ".git")):
            self._git("init")
            self._setcommitter()

    def gitignore(self, names):
        if not os.path.exists(os.path.join(self.path, ".gitignore")):
            with open(os.path.join(self.path, ".gitignore"), "w") as writer:
                for name in names:
                    writer.write("%s\n" % name)

    def add(self, *filenames):
        self._git("add", *filenames)

    def commit(self, message, was_retry=False):
        success = self._git("commit", "-m", message)
        if not success and not was_retry:
            self._setcommitter()
            self.commit(message, True)

    def _setcommitter(self):
        if not self._git("config", "user.name"):
            self._git("config", "user.name", "treeoflife-autocommit")
        if not self._git("config", "user.email"):
            self._git("config", "user.email", "treeoflife@localhost")

    def _git(self, *args):
        result = subprocess.call(
                [self.binary] + list(args), cwd=self.path)
        return result == 0


class SavingInterface(CommandInterface):
    def __init__(self, directory, main_file,
            config_file="config.json",
            autosave_template="_{main_file}_autosave_{time}",
            backup_template="_{main_file}_backup_{time}", **kw):
        super(SavingInterface, self).__init__(**kw)

        if directory is None:
            self.save_dir = None
            return
        self.save_dir = os.path.realpath(os.path.expanduser(directory))
        self.main_file = main_file
        self.save_file = os.path.join(self.save_dir, main_file)
        self.config_file = os.path.join(self.save_dir, config_file)
        self.autosave_file = os.path.join(self.save_dir, autosave_template)
        self.backup_file = os.path.join(self.save_dir, backup_template)
        self.timeformat = "%A %B %d %H:%M:%S %Y"

        self.last_auto_save = None
        self.last_backup_save = None
        self.last_full_save = None

        self.autosave_id = time.time()
        self.autosave_minutes = 5
        self.backup_minutes = 30

        self.git = Git(self.save_dir)

    def _load_file(self, filename, callback):
        try:
            reader = open(os.path.realpath(filename), "r")
        except IOError:
            return None
        else:
            return callback(reader)

    def load(self):
        if self.save_dir is None:
            return
        config = self._load_file(self.config_file, json.load)
        if config:
            self.config = config
        self._load_file(self.save_file, partial(self.deserialize, "file"))

    def full_save(self):
        if self.save_dir is None:
            return
        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

        self.git.init()

        with open(self.config_file, "w") as writer:
            json.dump(self.config, writer, sort_keys=True,
                    indent=4)
        self.git.add(self.config_file)

        with open(self.save_file, "w") as writer:
            self.serialize("file", writer)
        self.git.add(self.save_file)

        self.git.gitignore(["_*"])
        self.git.add(".gitignore")

        self.git.commit("Full save %s" %
                datetime.datetime.now().strftime(self.timeformat))
        self.last_full_save = datetime.datetime.now()

    def auto_save(self):
        if self.save_dir is None:
            return
        self._special_save(self.autosave_file, self.autosave_id,
                self.autosave_minutes, "last_auto_save")

    def _special_save(self, name_format, time, freq, lastname):
        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

        last = getattr(self, lastname)
        if last and datetime.datetime.now() < last + datetime.timedelta(
                                                        minutes=freq):
            return

        filename = name_format.format(main_file=self.main_file, time=int(time))
        with open(filename, "w") as writer:
            self.serialize("file", writer)
        setattr(self, lastname, datetime.datetime.now())



#!/usr/bin/env python
from setuptools import setup, find_packages

import platform
if platform.system() == "Windows":
    mainscript = "bin/treeoflife-server.py"
else:
    mainscript = "bin/treeoflife-server"

vers = "0.0.2"

setup(
        name="treeoflife",
        description="A Tree of your Life (end-user application)",
        version=vers,
        packages=["treeoflife"],
        license='MIT',
        author="lahwran",
        author_email="lahwran0@gmail.com",
        url="https://github.com/lahwran/tree-of-life",
        download_url="https://github.com/lahwran/tree-of-life/tarball/" + vers,
        scripts=[mainscript],
        install_requires=[
            "twisted",
            "parsley",
            "pytest",
            "pep8",
            "txws",
            "raven",
            "sphinx",
        ],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Web Environment',
            'Environment :: MacOS X',
            'Framework :: Twisted',
            'Intended Audience :: End Users/Desktop',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Operating System :: MacOS :: MacOS X',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 2 :: Only',
            'Programming Language :: Python :: Implementation :: PyPy',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Objective C',
            'Programming Language :: JavaScript',
            'Topic :: Office/Business :: Scheduling',
            'Topic :: Software Development :: Bug Tracking'
        ]
)



from __future__ import unicode_literals, print_function

from datetime import timedelta, datetime
import weakref

import pytest
from twisted.internet.task import Clock
from twisted.internet.defer import Deferred

from treeoflife.nodes.node import TreeRootNode, Node
from treeoflife import alarms
from treeoflife.test.util import FakeNodeCreator
from treeoflife.tracker import Tracker
from treeoflife.util import _monitor


@pytest.fixture
def reactortime(monkeypatch):
    clock = Clock()
    return clock

#-------------------------------------------------------#
#                   utility classes                     #
#-------------------------------------------------------#


class _PresetTracker(Tracker):
    def __init__(self, roottype, nodetype):
        self.nodecreator = FakeNodeCreator(nodetype)
        self.roottype = roottype
        self.root = roottype(self, self.nodecreator)
        self.make_skeleton = False
        self.alarms = []


class DummyTracker(_PresetTracker):
    def _set_alarm(self, alarm):
        self.alarms.append(alarm)


class DummyTimer(object):
    def __init__(self):
        self.is_active = True

    def cancel(self):
        self.is_active = False

    def active(self):
        return self.is_active


class DummyAlarmTracker(Tracker):
    def _set_alarm(self, callback=None, delta=None, date=None):
        deferred = Deferred()
        timer = DummyTimer()
        alarm = {
            "deferred": deferred,
            "timer": timer,
            "data": {
                "callback": callback,
                "delta": delta,
                "date": date
            }
        }
        self.alarms.append(alarm)
        return timer, deferred


class SuperDummyTracker(_PresetTracker):
    @property
    def _set_alarm(self):
        raise AttributeError("I don't exist!")


class MixedTracker(alarms.TrackerMixin, _PresetTracker):
    def __init__(self, *a, **kw):
        self._reactor = kw.pop("reactor")
        _PresetTracker.__init__(self, *a, **kw)


class AlarmCollectorRoot(TreeRootNode):
    def __init__(self, tracker, nodecreator):
        TreeRootNode.__init__(self, tracker, nodecreator)
        self.alarms = []

    def _add_alarm(self, alarm):
        self.alarms.append(alarm)


class MixedNode(Node, alarms.NodeMixin):
    def __init__(self, *args, **kw):
        Node.__init__(self, *args, **kw)
        alarms.NodeMixin.__init__(self)


class DummyAlarm(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.root = None
        self.is_live = False

    def _node_ready(self, root):
        self.root = root

    def _go_live(self):
        self.is_live = True


class MixedRoot(alarms.RootMixin, TreeRootNode):
    def __init__(self, *a, **kw):
        alarms.RootMixin.__init__(self)
        TreeRootNode.__init__(self, *a, **kw)


class InitAlarmNode(MixedNode):
    def __init__(self, *a, **kw):
        self.func_alarm = self.alarm(self.func)
        assert self.func_alarm.root is None

        MixedNode.__init__(self, *a, **kw)

    def func(self):
        pass

#-------------------------------------------------------#
#                        tests                          #
#-------------------------------------------------------#


class TestNodeMixin(object):
    def test_already_registered(self, monkeypatch):
        tracker = DummyTracker(AlarmCollectorRoot, MixedNode)
        node = MixedNode("node", None, tracker.root)
        tracker.root.addchild(node)
        assert set(node.alarms) == set()

        monkeypatch.setattr(alarms, "Alarm", DummyAlarm)

        argsentinel = object()
        alarm = node.alarm(argsentinel, 1, 2, 3, herp="derp")
        assert alarm.args == (node, argsentinel, 1, 2, 3)
        assert alarm.kwargs == {"herp": "derp"}
        assert alarm.root is node.root
        assert node.root is tracker.root
        assert tracker.root.alarms == [alarm]

    def test_not_registered(self, monkeypatch):
        tracker = DummyTracker(AlarmCollectorRoot, MixedNode)
        node = MixedNode("node", None, None)
        assert set(node.alarms) == set()

        monkeypatch.setattr(alarms, "Alarm", DummyAlarm)

        argsentinel = object()
        alarm = node.alarm(argsentinel, 1, 2, 3, herp="derp")
        assert alarm.args == (node, argsentinel, 1, 2, 3)
        assert alarm.kwargs == {"herp": "derp"}
        assert alarm.root is None
        assert tracker.root.alarms == []

        tracker.root.addchild(node)

        assert alarm.root is node.root
        assert node.root is tracker.root
        assert tracker.root.alarms == [alarm]


class TestRootMixin(object):
    def test_already_loaded(self):
        tracker = DummyTracker(MixedRoot, MixedNode)
        root = tracker.root
        assert set(root.alarms) == set()
        assert root._alarm_ready

        alarm = DummyAlarm()
        assert not alarm.is_live

        root._add_alarm(alarm)

        assert alarm.is_live

    def test_load_alarm(self, monkeypatch):
        monkeypatch.setattr(alarms, "Alarm", DummyAlarm)
        tracker = DummyTracker(MixedRoot, InitAlarmNode)

        tracker.deserialize("str", u"node: test")

        node = tracker.root.children.next_neighbor
        assert node.func_alarm.root is node.root
        assert node.func_alarm.is_live

    def test_no_support_noop(self, monkeypatch):
        monkeypatch.setattr(alarms, "Alarm", DummyAlarm)
        tracker = SuperDummyTracker(MixedRoot, InitAlarmNode)
        tracker.deserialize("str", u"node: test")

        node = tracker.root.children.next_neighbor
        assert node.func_alarm.root is node.root
        assert not node.func_alarm.is_live

    @pytest.mark.weakref
    def test_thrown_away(self, monkeypatch):
        tracker = DummyTracker(MixedRoot, MixedNode)
        root = tracker.root
        assert set(root.alarms) == set()
        assert root._alarm_ready

        alarm = DummyAlarm()
        assert not alarm.is_live

        root._add_alarm(alarm)
        assert set(root.alarms) == set([alarm])
        ref = weakref.ref(alarm)
        del alarm
        import gc; gc.collect()
        assert not ref()

        assert set(root.alarms) == set()


class TestSetAlarm(object):
    def test_calling(self, reactortime):
        tracker = MixedTracker(TreeRootNode, Node,
                reactor=reactortime)

        calls = []

        def callback(passed_tracker):
            assert passed_tracker is tracker
            calls.append(1)

        # anchor retains a reference to the deferred
        anchor = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=30))

        assert calls == []
        reactortime.advance(29.9)
        assert calls == []
        reactortime.advance(0.2)
        assert calls == [1]

    def test_delta_calculation(self, reactortime, setdt):
        tracker = MixedTracker(TreeRootNode, Node,
                reactor=reactortime)
        setdt(2012, 12, 21)
        target = datetime(2012, 12, 21, 23, 59, 59)
        day_seconds = 60 * 60 * 24

        calls = []

        def callback(passed_tracker):
            assert passed_tracker is tracker
            calls.append(1)

        # anchor retains a reference to the deferred
        anchor = tracker._set_alarm(callback=callback, date=target)

        assert calls == []
        reactortime.advance(day_seconds - 1.1)
        assert calls == []
        reactortime.advance(0.2)
        assert calls == [1]

    def test_negative_delta(self, reactortime, setdt):
        tracker = MixedTracker(TreeRootNode, Node,
                reactor=reactortime)
        setdt(2012, 12, 21)
        target = datetime(2012, 12, 20)
        day_seconds = 60 * 60 * 24

        with pytest.raises(ValueError):
            tracker._set_alarm(date=target)

    @pytest.mark.weakref
    def test_deleted_root(self, reactortime):
        tracker = MixedTracker(TreeRootNode, Node,
                reactor=reactortime)

        def callback(passed_tracker):  # pragma: no cover
            assert False, "shouldn't have been called"

        # anchor retains a reference to the deferred
        anchor_1 = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=30))
        anchor_2 = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=60))

        # break the reference loop before throwing it away
        del tracker.root._root
        ref = weakref.ref(tracker.root)
        tracker.deserialize("str", u"node: somenode")
        import gc; gc.collect()
        assert not ref()

        root = tracker.root
        # to force an error should it not detect it via weakref
        del tracker.root

        reactortime.advance(31)
        assert not reactortime.getDelayedCalls()

    @pytest.mark.weakref
    def test_deleted_tracker(self, reactortime):
        tracker = MixedTracker(TreeRootNode, Node,
                reactor=reactortime)

        def callback(passed_tracker):  # pragma: no cover
            assert False, "shouldn't have been called"

        # anchor retains a reference to the deferred
        anchor_1 = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=30))
        anchor_2 = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=60))

        # break the reference loop before throwing it away
        root_anchor = tracker.root
        vars(root_anchor).clear()

        tracker.deserialize("str", u"node: somenode")

        # to force an error should it not detect it via weakref
        vars(tracker).clear()
        ref = weakref.ref(tracker)
        del tracker
        import gc; gc.collect()
        assert not ref()

        reactortime.advance(31)
        assert not reactortime.getDelayedCalls()

    @pytest.mark.weakref
    def test_abandoned_deferred(self, reactortime):
        tracker = MixedTracker(TreeRootNode, Node,
                reactor=reactortime)

        def callback(passed_tracker):  # pragma: no cover
            assert False, "shouldn't have been called"

        # anchor retains a reference to the deferred
        timer, deferred = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=30))
        timer_2, deferred_2 = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=60))

        ref = weakref.ref(deferred)
        del deferred
        import gc; gc.collect()
        assert not ref()

        reactortime.advance(31)
        assert reactortime.getDelayedCalls() == [timer_2]

    def test_deterministic_fallback(self, reactortime):
        tracker = MixedTracker(TreeRootNode, Node,
                reactor=reactortime)

        def callback(passed_tracker):  # pragma: no cover
            assert False, "shouldn't have been called"

        # anchor retains a reference to the deferred
        anchor_1 = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=30))
        anchor_2 = tracker._set_alarm(callback=callback,
                delta=timedelta(seconds=60))

        # make sure to hold the reference
        old_root = tracker.root

        tracker.deserialize("str", u"node: somenode")
        import gc; gc.collect()

        reactortime.advance(31)
        assert not reactortime.getDelayedCalls()


class TestAlarm(object):
    def test_delta(self, setdt, reactortime):
        setdt(2012, 12, 29, 4, 33)

        class SomeNode(MixedNode):
            def __init__(self, *a, **kw):
                MixedNode.__init__(self, *a, **kw)
                self.the_alarm = self.alarm(self.callback,
                        delta=timedelta(seconds=30))
                self.callback_called = False

            def callback(self):
                self.callback_called = True

        tracker = MixedTracker(MixedRoot, SomeNode,
                reactor=reactortime)
        node = tracker.root.createchild(u"some", u"node")
        assert not node.callback_called

        reactortime.advance(31)
        assert node.callback_called

    def test_alter(self, setdt, reactortime):
        setdt(2012, 12, 29, 5, 41)

        class SomeNode(MixedNode):
            def __init__(self, *a, **kw):
                MixedNode.__init__(self, *a, **kw)
                self.the_alarm = self.alarm(self.callback,
                        delta=timedelta(seconds=30))
                self.callback_called = False

            def callback(self):
                self.callback_called = True

        tracker = MixedTracker(MixedRoot, SomeNode,
                reactor=reactortime)
        node = tracker.root.createchild(u"some", u"node")
        assert not node.callback_called

        reactortime.advance(29)
        setdt(2012, 12, 29, 5, 41, 29)
        assert not node.callback_called

        node.the_alarm.date = node.the_alarm.date + timedelta(seconds=30)
        reactortime.advance(2)
        assert not node.callback_called

        reactortime.advance(30)
        assert node.callback_called

    def test_in_past(self, setdt):
        setdt(2012, 12, 29, 5, 44)

        class SomeNode(MixedNode):
            def __init__(self, *a, **kw):
                MixedNode.__init__(self, *a, **kw)
                self.the_alarm = self.alarm(self.callback,
                        delta=timedelta(seconds=-30))
                self.callback_called = False

            def callback(self):
                self.callback_called = True

        tracker = MixedTracker(MixedRoot, SomeNode,
                reactor=reactortime)
        node = tracker.root.createchild(u"some", u"node")
        assert node.callback_called



from __future__ import unicode_literals, print_function

import json

from twisted.internet.task import Clock
import pytest
import pprint

from treeoflife.main import JSONProtocol, RemoteInterface
from treeoflife import editor_launch
from treeoflife.test import util

# TODO: need to monkeypatch tempfile() to not clutter os temp directory


class NoIOProtocol(JSONProtocol):
    def __init__(self, *a, **kw):
        self.allowed = kw.pop("allowed")
        JSONProtocol.__init__(self, *a, **kw)
        self._is_transient_connection = True
        self.sent_messages = []
        self.accepted_messages = []
        self.do_raise = True
        self.transport = FakeTransport()

    def accept_messages(self, *messages):
        self.accepted_messages.extend(messages)
        assert self.sent_messages == self.accepted_messages

    def update(self):
        if "update" in self.allowed:
            return JSONProtocol.update(self)

    def update_all(self):
        if "update_all" in self.allowed:
            return JSONProtocol.update_all(self)

    def update_editor_running(self):
        if "update_editor_running" in self.allowed:
            return JSONProtocol.update_editor_running(self)

    def sendmessage(self, message):
        self.sent_messages.append(message)

    def receive(self, **kw):
        self.line_received(json.dumps(kw))

    def capture_error(self, e, message=None):
        if self.do_raise:
            raise
        else:
            JSONProtocol.capture_error(self, e, message)


@pytest.fixture
def tpc():
    class Config(object):
        editor = "test-editor"
        port = 12345

    clock = Clock()
    tracker = RemoteInterface(Config, None, None, None,
            reactor=clock)
    protocol = NoIOProtocol(tracker, clock, allowed={"update_editor_running"})
    tracker.listeners.append(protocol)

    assert not tracker.edit_session
    return tracker, protocol, clock


@editor_launch.editor_types.add("test-editor")
class EditorForTesting(editor_launch._TerminalLauncher):
    pass


class FakeTransport(object):
    def __init__(self):
        self.connected = True
        self.disconnecting = False

    def loseConnection(self):
        self.connected = False


def test_edit(tpc):
    tracker, protocol, clock = tpc

    protocol.receive(command="edit")

    protocol.accept_messages(
        {u"display": False},
        {u"editor_running": True}
    )
    assert tracker.edit_session.editor.command

    with open(tracker.edit_session.editor.tmp, "a") as writer:
        writer.write("\ncomment: \xfcherp derp\n".encode("utf-8"))

    temp_protocol = NoIOProtocol(tracker, clock, allowed=set())
    temp_protocol.line_received(tracker.edit_session.editor.json_data)
    assert not tracker.root.find(u"\xfcherp derp").first()
    protocol.accept_messages()
    assert not temp_protocol.transport.connected

    clock.advance(1)
    assert tracker.root.find(u"\xfcherp derp").one()
    assert not temp_protocol.sent_messages
    protocol.accept_messages(
        {u"editor_running": False},
        {u"display": True}
    )


def test_edit_nohide(tpc, monkeypatch):
    tracker, protocol, clock = tpc

    monkeypatch.setattr(EditorForTesting, "hide", False)

    protocol.receive(command="edit")

    protocol.accept_messages({u"editor_running": True})
    assert tracker.edit_session.editor.command

    with open(tracker.edit_session.editor.tmp, "a") as writer:
        writer.write("\ncomment: \xfcherp derp\n".encode("utf-8"))

    temp_protocol = NoIOProtocol(tracker, clock, allowed=set())

    temp_protocol.line_received(tracker.edit_session.editor.json_data)
    protocol.accept_messages()
    temp_protocol.accept_messages()
    assert not temp_protocol.transport.connected

    clock.advance(1)
    protocol.accept_messages({u"editor_running": False})


def test_edit_nochange(tpc, monkeypatch):
    tracker, protocol, clock = tpc
    root = tracker.root
    monkeypatch.setattr(EditorForTesting, "hide", False)
    protocol.receive(command="edit")

    protocol.accept_messages({u"editor_running": True})
    assert tracker.edit_session.editor.command

    temp_protocol = NoIOProtocol(tracker, clock, allowed=set())

    temp_protocol.line_received(tracker.edit_session.editor.json_data)
    protocol.accept_messages()
    temp_protocol.accept_messages()
    assert not temp_protocol.transport.connected

    clock.advance(1)
    protocol.accept_messages({u"editor_running": False})
    assert tracker.root is root


def test_edit_error(tpc):
    tracker, protocol, clock = tpc
    tracker.root.createchild("task", "test")
    protocol.receive(command="edit")

    assert protocol.sent_messages == [
        {u"display": False},
        {u"editor_running": True}
    ]
    protocol.sent_messages = []
    editor = tracker.edit_session.editor
    assert editor.command
    identifier = editor.identifier
    tmp = editor.tmp
    root = tracker.root

    with open(editor.tmp, "r") as reader:
        data = reader.read()
    with open(editor.tmp, "w") as writer:
        writer.write("\nday: dasfadsf\xfcherp derp\n".encode("utf-8"))
        writer.write(data)

    temp_protocol = NoIOProtocol(tracker, clock, allowed=set())
    temp_protocol.transport = FakeTransport()

    temp_protocol.line_received(tracker.edit_session.editor.json_data)
    assert not tracker.root.find(u"dasfadsf\xfcherp derp").first()
    assert not protocol.sent_messages
    assert not temp_protocol.transport.connected
    assert tracker.edit_session.editor.identifier == identifier
    assert tracker.edit_session.editor.tmp == tmp
    assert tracker.root is root

    protocol.do_raise = False
    clock.advance(1)
    assert not tracker.root.find(u"dasfadsf\xfcherp derp").first()
    assert tracker.edit_session.editor.identifier != identifier
    assert len(tracker.edit_session.editor.filenames) == 2
    assert tracker.edit_session.editor.error
    assert tracker.edit_session.editor.error_tmp
    assert tracker.root.find("test").one()
    assert not temp_protocol.sent_messages
    assert len(protocol.sent_messages) == 1
    assert protocol.sent_messages[0].keys() == ["error"]
    assert "LoadError" in protocol.sent_messages[0]["error"]


def test_edit_embedded():
    class Config(object):
        editor = "embedded"
        port = 12345

    clock = Clock()
    tracker = RemoteInterface(Config, None, None, None,
            reactor=clock)
    protocol = NoIOProtocol(tracker, clock, allowed={"update_editor_running"})
    tracker.listeners.append(protocol)

    protocol.receive(command="edit")
    protocol.accept_messages(
        {"embedded_edit": util.any},
        {"editor_running": True}
    )

    msg = dict(protocol.sent_messages[0]["embedded_edit"])
    assert msg["identifier"] == tracker.edit_session.editor.identifier
    data = msg["data"]
    assert data
    assert not msg["error"]

    data += "\ncomment: \xfcherp derp\n"

    response_message = json.dumps({
        "embedded_editor_finished": {
            "identifier": msg["identifier"],
            "data": data
        }
    })
    protocol.accept_messages()
    protocol.line_received(response_message)
    protocol.accept_messages(
        {"editor_running": False},
        {"embedded_edit": None}
    )
    assert tracker.root.find(u"\xfcherp derp").one()
    assert not tracker.edit_session


def test_receive_long_message():
    # long messages are used to transfer the entire life file.
    # they're allowed, darn you default twisted linereceiver

    length = 5000000  # 5MB
    called = []

    class Derp(NoIOProtocol):
        def message_huge(self, data):
            assert data == " " * length
            called.append(True)

    class Config(object):  # not really needed I think
        editor = "test-editor"
        port = 12345

    clock = Clock()
    tracker = RemoteInterface(Config, None, None, None,
            reactor=clock)
    protocol = Derp(tracker, clock, allowed={"update_editor_running"})

    protocol.dataReceived(json.dumps({"huge": " " * length}) + "\n")
    assert called == [True]



from __future__ import unicode_literals, print_function

import logging
logger = logging.getLogger(__name__)


class ListIntegrityError(Exception):
    pass


class ErrorContext(object):
    def __init__(self):
        self.line = None


class LoadError(Exception):
    def __init__(self, message):
        self._message = message

    @property
    def message(self):
        return str(self)

    def __str__(self):
        try:
            result = ""
            if getattr(self, "error_context", False):
                line = getattr(self.error_context, "line", None)
                if line is not None:
                    line = int(line) + 1  # zero indexed to human indexed
                    result = "At line %d: " % line
                else:
                    result = "At unknown line: "
            result += self._message
            return result
        except Exception as e:
            logger.exception("Error formatting error")
            return self._message


class CantStartNodeError(LoadError):
    pass


class InvalidInputError(Exception):
    pass



from __future__ import unicode_literals, print_function

import pytest

from treeoflife import userinterface
from treeoflife import exceptions
from treeoflife.tracker import Tracker


class Test_makenode(object):
    def test_makenode_indent(self, monkeypatch):
        result = 1, False, u"abcde", u"\xfcnode_type", u"\xfctext"
        monkeypatch.setattr(userinterface, "parse_line", lambda string: result)

        with pytest.raises(exceptions.InvalidInputError):
            userinterface._makenode(u"\xfcsentinel")

    def test_makenode_metadata(self, monkeypatch):
        result = 0, True, u"abcde", u"node_type", u"text"
        monkeypatch.setattr(userinterface, "parse_line", lambda string: result)

        with pytest.raises(exceptions.InvalidInputError):
            userinterface._makenode(u"\xfcsentinel")

    def test_makenode(self):
        assert userinterface._makenode(u"\xfctask: \xfctest") == (
                u"\xfctask", u"\xfctest")


class TestGenerateListing(object):
    def stuff(self):
        tracker = Tracker(skeleton=False)
        tracker.root.createchild("_gennode", "herp")
        derp = tracker.root.createchild("_gennode", "derp")
        active = derp.createchild("task", "honk")
        return tracker, derp, active

    def test_simple(self):
        tracker, derp, active = self.stuff()

        result = userinterface.generate_listing(active, tracker.root)

        assert result == [
            "  life",
            "      _gennode: herp",
            "      _gennode: derp",
            ">         task: honk"
        ]



from __future__ import unicode_literals, print_function

import datetime

from treeoflife.parseutil import Grammar
from treeoflife.nodes.node import Node, Option, BooleanOption, nodecreator
from treeoflife import timefmt


class ActiveMarker(BooleanOption):
    name = "active"

    def set(self, node, value):
        node.root.activate(node)

    def get(self, node):
        return (node is node.root.active_node), None


class _FinishGrammar(Grammar):
    grammar = """
    finished = (time.timedelta:td wss 'after' wss time.datetime:dt
                        -> ("both", (td, dt))
               |time.datetime:dt -> ("finished", dt)
               )
    """
    bindings = {
        "time": timefmt.TimeGrammar
    }


class StartedOption(timefmt.DatetimeOption):
    def set(self, node, value):
        if value is None:
            node.started = datetime.datetime.now()
        else:
            super(StartedOption, self).set(node, value)

    def get(self, node):
        if getattr(node, "finished", None) is not None:
            return False, None
        else:
            return super(StartedOption, self).get(node)


class FinishedOption(object):
    name = "finished"

    def set(self, node, value):
        if value is None:
            node.finished = datetime.datetime.now()
            return

        kind, value = _FinishGrammar(value).finished()
        if kind == "finished":
            node.finished = value
        elif kind == "both":
            delta, started = value
            node.finished = started + delta
            node.started = started

    def get(self, node):
        started = getattr(node, "started", None)
        finished = getattr(node, "finished", None)
        if finished is None:
            return False, None
        elif started is None:
            return True, timefmt.datetime_to_str(finished)
        else:
            delta = finished - started
            return True, "%s after %s" % (
                        timefmt.timedelta_to_str(delta),
                        timefmt.datetime_to_str(started) )


@nodecreator("worked on")
class BaseTask(Node):
    options = (
        StartedOption("started"),
        FinishedOption(),
        ActiveMarker()
    )

    def __init__(self, *args, **kwargs):
        super(BaseTask, self).__init__(*args, **kwargs)

        self.started = None
        self.finished = None

    def start(self):
        if self.started:
            return
            # wtf do we do now?
            raise Exception("fixme, need to do something about restarts")
        self.started = datetime.datetime.now()

    def finish(self):
        self.finished = datetime.datetime.now()

    def unfinish(self):
        self.finished = None
        return True

    @property
    def can_activate(self):
        return self.finished is None

    def ui_dictify(self, result=None):
        if result is None:
            result = {}

        result["started"] = bool(self.started)
        result["finished"] = bool(self.finished)
        result["active_id"] = self.active_id
        return super(BaseTask, self).ui_dictify(result)

    def search_tags(self):
        tags = Node.search_tags(self)
        if self.finished:
            tags.add("finished")
        else:
            tags.add("unfinished")
        if self.started:
            tags.add("started")
        else:
            tags.add("unstarted")
        if self is self.root.active_node:
            tags.add("active")
        else:
            tags.add("inactive")

        return tags


@nodecreator("task")
@nodecreator("question")
@nodecreator("problem")
@nodecreator("project")
@nodecreator("bug")
@nodecreator("feature")
class Task(BaseTask):
    multiline = True


@nodecreator("category")
class Category(Node):
    # should be passthrough
    pass


@nodecreator("event")
class Event(BaseTask):
    multiline = True
    options = (
        timefmt.DatetimeOption("when"),
        # need "where" option maybe
    )



from __future__ import unicode_literals, print_function

from treeoflife.userinterface import Event
from treeoflife.tracker import Tracker
from treeoflife import navigation
from treeoflife.nodes import days


def test_createauto_todo_integration():
    tracker = Tracker(skeleton=False)
    days = tracker.root.createchild("days")
    day = days.createchild("day", "today")
    tracker.root.activate(day)
    tracker.root.createchild("todo bucket", None)
    event = Event(None, tracker.root, "createauto",
        "todo: \xfctest",
        tracker
    )
    assert not tracker.root.find("todo bucket > todo: \xfctest").first()
    event._inject(navigation.createauto)
    assert tracker.root.find("todo bucket > todo: \xfctest").one()


def test_createauto_activatefirst(setdt):
    setdt(2014, 2, 19, 12)
    tracker = Tracker(skeleton=False)

    tracker.deserialize("str",
        "task: something\n"
        "days\n"
        "    day: today\n"
        "        @active\n"
        "    day: tomorrow\n"
        "    day: September 20, 2014\n"
    )

    day = tracker.root.find("today").one()
    assert tracker.root.active_node is day

    navigation.createauto("task: test", tracker.root)
    node = day.find("test").one()
    assert tracker.root.active_node is node

    navigation.createauto("today", tracker.root)
    assert tracker.root.active_node is day

    navigation.createauto("task: test", tracker.root)
    assert tracker.root.active_node is node
